@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -233,11 +233,39 @@ join_threads(void)
 	} while (waited);
 }
 
+void
+join_detached_threads(void)
+{
+	struct winthread *w;
+	int waited;
+
+	do {
+		waited = 0;
+		EnterCriticalSection(&winthread_cs);
+		for (w = winthreads; w; w = w->next) {
+			if ((w->flags & (DETACHED | WAITING)) == DETACHED) {
+				w->flags |= WAITING;
+				LeaveCriticalSection(&winthread_cs);
+				WaitForSingleObject(w->hdl, INFINITE);
+				CloseHandle(w->hdl);
+				rm_winthread(w);
+				waited = 1;
+				EnterCriticalSection(&winthread_cs);
+				break;
+			}
+		}
+		LeaveCriticalSection(&winthread_cs);
+	} while (waited);
+}
+
 int
 MT_create_thread(MT_Id *t, void (*f) (void *), void *arg, enum MT_thr_detach d)
 {
 	struct winthread *w = malloc(sizeof(*w));
 
+	if (w == NULL)
+		return -1;
+
 	if (winthread_cs_init == 0) {
 		/* we only get here before any threads are created,
 		 * and this is the only time that winthread_cs_init is
@@ -392,7 +420,6 @@ pthread_sema_down(pthread_sema_t *s)
 
 #else  /* !defined(HAVE_PTHREAD_H) && defined(_MSC_VER) */
 
-#ifdef HAVE_PTHREAD_SIGMASK
 static struct posthread {
 	struct posthread *next;
 	pthread_t tid;
@@ -427,6 +454,7 @@ find_posthread(pthread_t tid)
 }
 #endif
 
+#ifdef HAVE_PTHREAD_SIGMASK
 static void
 MT_thread_sigmask(sigset_t * new_mask, sigset_t * orig_mask)
 {
@@ -448,14 +476,6 @@ rm_posthread_locked(struct posthread *p)
 		*pp = p->next;
 }
 
-static void
-rm_posthread(struct posthread *p)
-{
-	pthread_mutex_lock(&posthread_lock);
-	rm_posthread_locked(p);
-	pthread_mutex_unlock(&posthread_lock);
-}
-
 static void
 thread_starter(void *arg)
 {
@@ -497,6 +517,25 @@ join_threads(void)
 	pthread_mutex_unlock(&posthread_lock);
 }
 
+void
+join_detached_threads(void)
+{
+	struct posthread *p;
+	pthread_t tid;
+
+	pthread_mutex_lock(&posthread_lock);
+	while (posthreads) {
+		p = posthreads;
+		posthreads = p->next;
+		tid = p->tid;
+		free(p);
+		pthread_mutex_unlock(&posthread_lock);
+		pthread_join(tid, NULL);
+		pthread_mutex_lock(&posthread_lock);
+	}
+	pthread_mutex_unlock(&posthread_lock);
+}
+
 int
 MT_create_thread(MT_Id *t, void (*f) (void *), void *arg, enum MT_thr_detach d)
 {
@@ -518,13 +557,15 @@ MT_create_thread(MT_Id *t, void (*f) (void *), void *arg, enum MT_thr_detach d)
 	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
 	if (d == MT_THR_DETACHED) {
 		p = malloc(sizeof(struct posthread));
+		if (p == NULL) {
+#ifdef HAVE_PTHREAD_SIGMASK
+			MT_thread_sigmask(&orig_mask, NULL);
+#endif
+			return -1;
+		}
 		p->func = f;
 		p->arg = arg;
 		p->exited = 0;
-		pthread_mutex_lock(&posthread_lock);
-		p->next = posthreads;
-		posthreads = p;
-		pthread_mutex_unlock(&posthread_lock);
 		f = thread_starter;
 		arg = p;
 		newtp = &p->tid;
@@ -539,8 +580,13 @@ MT_create_thread(MT_Id *t, void (*f) (void *), void *arg, enum MT_thr_detach d)
 #else
 		*t = (MT_Id) (((size_t) *newtp) + 1);	/* use pthread-id + 1 */
 #endif
+		if (p) {
+			pthread_mutex_lock(&posthread_lock);
+			p->next = posthreads;
+			posthreads = p;
+			pthread_mutex_unlock(&posthread_lock);
+		}
 	} else if (p) {
-		rm_posthread(p);
 		free(p);
 	}
 #ifdef HAVE_PTHREAD_SIGMASK
@@ -765,6 +811,9 @@ MT_check_nr_cores_(void)
 		lng t0, t1;
 		MT_Id *threads = malloc(sizeof(MT_Id) * curr);
 
+		if (threads == NULL)
+			break;
+
 		t0 = GDKusec();
 		for (i = 0; i < curr; i++)
 			MT_create_thread(threads + i, smp_thread, NULL, MT_THR_JOINABLE);
@@ -831,46 +880,7 @@ MT_check_nr_cores(void)
 lng
 GDKusec(void)
 {
-	/* Return the time in microseconds since an epoch.  The epoch
-	 * is roughly the time this program started. */
-#ifdef _MSC_VER
-	static LARGE_INTEGER freq, start;	/* automatically initialized to 0 */
-	LARGE_INTEGER ctr;
-
-	if (start.QuadPart == 0 &&
-	    (!QueryPerformanceFrequency(&freq) ||
-	     !QueryPerformanceCounter(&start)))
-		start.QuadPart = -1;
-	if (start.QuadPart > 0) {
-		QueryPerformanceCounter(&ctr);
-		return (lng) (((ctr.QuadPart - start.QuadPart) * 1000000) / freq.QuadPart);
-	}
-#endif
-#ifdef HAVE_GETTIMEOFDAY
-	{
-		static struct timeval tpbase;	/* automatically initialized to 0 */
-		struct timeval tp;
-
-		if (tpbase.tv_sec == 0)
-			gettimeofday(&tpbase, NULL);
-		gettimeofday(&tp, NULL);
-		tp.tv_sec -= tpbase.tv_sec;
-		return (lng) tp.tv_sec * 1000000 + (lng) tp.tv_usec;
-	}
-#else
-#ifdef HAVE_FTIME
-	{
-		static struct timeb tbbase;	/* automatically initialized to 0 */
-		struct timeb tb;
-
-		if (tbbase.time == 0)
-			ftime(&tbbase);
-		ftime(&tb);
-		tb.time -= tbbase.time;
-		return (lng) tb.time * 1000000 + (lng) tb.millitm * 1000;
-	}
-#endif
-#endif
+return 0; // hahahahahahah
 }
 
 
