@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -42,7 +42,6 @@
 /* (author) M.L. Kersten */
 #include "monetdb_config.h"
 #include "mal_client.h"
-#include "mal_readline.h"
 #include "mal_import.h"
 #include "mal_parser.h"
 #include "mal_namespace.h"
@@ -50,15 +49,6 @@
 #include "mal_runtime.h"
 #include "mal_authorize.h"
 
-/*
- * This should be in src/mal/mal.h, as the function is implemented in
- * src/mal/mal.c; however, it cannot, as "Client" isn't known there ...
- * |-( For now, we move the prototype here, as it it only used here.
- * Maybe, we should consider also moving the implementation here...
- */
-
-static void freeClient(Client c);
-
 int MAL_MAXCLIENTS = 0;
 ClientRec *mal_clients;
 
@@ -80,7 +70,10 @@ MCinit(void)
 		maxclients = atoi(max_clients);
 	if (maxclients <= 0) {
 		maxclients = 64;
-		GDKsetenv("max_clients", "64");
+		if (GDKsetenv("max_clients", "64") != GDK_SUCCEED) {
+			showException(GDKout, MAL, "MCinit", "GDKsetenv failed");
+			mal_exit();
+		}
 	}
 
 	MAL_MAXCLIENTS =
@@ -88,7 +81,7 @@ MCinit(void)
 		/* client connections */ maxclients;
 	mal_clients = GDKzalloc(sizeof(ClientRec) * MAL_MAXCLIENTS);
 	if( mal_clients == NULL){
-		showException(GDKout, MAL, "MCinit",MAL_MALLOC_FAIL);
+		showException(GDKout, MAL, "MCinit", MAL_MALLOC_FAIL);
 		mal_exit();
 	}
 }
@@ -119,7 +112,7 @@ MCpopClientInput(Client c)
 	ClientInput *x = c->bak;
 	if (c->fdin) {
 		/* missing protection against closing stdin stream */
-		(void) bstream_destroy(c->fdin);
+		bstream_destroy(c->fdin);
 	}
 	GDKfree(c->prompt);
 	c->fdin = x->fdin;
@@ -135,6 +128,9 @@ static Client
 MCnewClient(void)
 {
 	Client c;
+	if (!mal_clients) {
+		return NULL;
+	}
 	MT_lock_set(&mal_contextLock);
 	if (mal_clients[CONSOLE].user && mal_clients[CONSOLE].mode == FINISHCLIENT) {
 		/*system shutdown in progress */
@@ -153,7 +149,7 @@ MCnewClient(void)
 		return NULL;
 	c->idx = (int) (c - mal_clients);
 #ifdef MAL_CLIENT_DEBUG
-	printf("New client created %d\n", (int) (c - mal_clients));
+	fprintf(stderr,"New client created %d\n", (int) (c - mal_clients));
 #endif
 	return c;
 }
@@ -181,7 +177,7 @@ void
 MCexitClient(Client c)
 {
 #ifdef MAL_CLIENT_DEBUG
-	printf("# Exit client %d\n", c->idx);
+	fprintf(stderr,"# Exit client %d\n", c->idx);
 #endif
 	finishSessionProfiler(c);
 	MPresetProfiler(c->fdout);
@@ -193,7 +189,7 @@ MCexitClient(Client c)
 		assert(c->bak == NULL);
 		if (c->fdin) {
 			/* missing protection against closing stdin stream */
-			(void) bstream_destroy(c->fdin);
+			bstream_destroy(c->fdin);
 		}
 		c->fdout = NULL;
 		c->fdin = NULL;
@@ -213,6 +209,9 @@ MCinitClientRecord(Client c, oid user, bstream *fin, stream *fout)
 	c->blkmode = 0;
 
 	c->fdin = fin ? fin : bstream_create(GDKin, 0);
+	if (!c->fdin) {
+		return NULL;
+	}
 	c->yycur = 0;
 	c->bak = NULL;
 
@@ -223,7 +222,9 @@ MCinitClientRecord(Client c, oid user, bstream *fin, stream *fout)
 	c->curprg = c->backup = 0;
 	c->glb = 0;
 
-	/* remove garbage from previous connection */
+	/* remove garbage from previous connection 
+	 * be aware, a user can introduce several modules 
+	 * that should be freed to avoid memory leaks */
 	if (c->nspace) {
 		freeModule(c->nspace);
 		c->nspace = 0;
@@ -237,26 +238,34 @@ MCinitClientRecord(Client c, oid user, bstream *fin, stream *fout)
 	c->stimeout = 0;
 	c->stage = 0;
 	c->itrace = 0;
-	c->debugOptimizer = c->debugScheduler = 0;
 	c->flags = 0;
 	c->errbuf = 0;
 
 	prompt = !fin ? GDKgetenv("monet_prompt") : PROMPT1;
 	c->prompt = GDKstrdup(prompt);
+	if (!c->prompt) {
+		return NULL;
+	}
 	c->promptlength = strlen(prompt);
 
 	c->actions = 0;
 	c->totaltime = 0;
-	/* create a recycler cache */
 	c->exception_buf_initialized = 0;
 	c->error_row = c->error_fld = c->error_msg = c->error_input = NULL;
 #ifndef HAVE_EMBEDDED /* no authentication in embedded mode */
 	{
 		str msg = AUTHgetUsername(&c->username, c);
 		if (msg)				/* shouldn't happen */
-			GDKfree(msg);
+			freeException(msg);
 	}
+#else
+	c->progress_callback = NULL;
+	c->progress_data = NULL;
+	MT_lock_init(&c->progress_lock, "progress_lock");
 #endif
+	c->blocksize = BLOCK;
+	c->protocol = PROTOCOL_9;
+	c->compute_column_widths = 0;
 	MT_sema_init(&c->s, 0, "Client->s");
 	return c;
 }
@@ -300,7 +309,7 @@ MCinitClientThread(Client c)
 	if (c->errbuf == NULL) {
 		char *n = GDKzalloc(GDKMAXERRLEN);
 		if ( n == NULL){
-			showException(GDKout, MAL, "initClientThread", "Failed to initialize client");
+			showException(GDKout, MAL, "initClientThread", MAL_MALLOC_FAIL);
 			return -1;
 		}
 		GDKsetbuf(n);
@@ -335,12 +344,9 @@ MCforkClient(Client father)
 		son->scenario = father->scenario;
 		if (son->prompt)
 			GDKfree(son->prompt);
-		son->prompt = GDKstrdup(father->prompt);
-		son->promptlength = strlen(father->prompt);
-		/* reuse the scopes wherever possible */
-		if (son->nspace == 0)
-			son->nspace = newModule(NULL, putName("child"));
-		son->nspace->outer = father->nspace->outer;
+		son->prompt = NULL;
+		son->promptlength = 0;
+		son->nspace = newModule(NULL, putName("child"));
 	}
 	return son;
 }
@@ -356,14 +362,14 @@ MCforkClient(Client father)
  * effects of sharing IO descriptors, also its children. Conversely, a
  * child can not close a parent.
  */
-void
+static void
 freeClient(Client c)
 {
 	Thread t = c->mythread;
 	c->mode = FINISHCLIENT;
 
 #ifdef MAL_CLIENT_DEBUG
-	printf("# Free client %d\n", c->idx);
+	fprintf(stderr,"# Free client %d\n", c->idx);
 #endif
 	MCexitClient(c);
 
@@ -395,13 +401,15 @@ freeClient(Client c)
 		c->username = 0;
 	}
 	c->mythread = 0;
-	GDKfree(c->glb);
-	c->glb = NULL;
+	if (c->glb) {
+		freeStack(c->glb);
+		c->glb = NULL;
+	}
 	if( c->error_row){
-		BBPdecref(c->error_row->batCacheid,TRUE);
-		BBPdecref(c->error_fld->batCacheid,TRUE);
-		BBPdecref(c->error_msg->batCacheid,TRUE);
-		BBPdecref(c->error_input->batCacheid,TRUE);
+		BBPrelease(c->error_row->batCacheid);
+		BBPrelease(c->error_fld->batCacheid);
+		BBPrelease(c->error_msg->batCacheid);
+		BBPrelease(c->error_input->batCacheid);
 		c->error_row = c->error_fld = c->error_msg = c->error_input = NULL;
 	}
 	if (t)
@@ -436,9 +444,12 @@ void
 MCstopClients(Client cntxt)
 {
 	Client c = mal_clients;
+	if (!mal_clients) {
+		return;
+	}
 	MT_lock_set(&mal_contextLock);
 	for (c = mal_clients + 1; c < mal_clients + MAL_MAXCLIENTS; c++) {
-		if (cntxt != c){
+		if (c && cntxt != c) {
 			if (c->mode == RUNCLIENT){
 				c->mode = FINISHCLIENT;
 			}
@@ -469,7 +480,7 @@ void
 MCcloseClient(Client c)
 {
 #ifdef MAL_DEBUG_CLIENT
-	printf("closeClient %d " OIDFMT "\n", (int) (c - mal_clients), c->user);
+	fprintf(stderr,"closeClient %d " OIDFMT "\n", (int) (c - mal_clients), c->user);
 #endif
 	/* free resources of a single thread */
 	if (!isAdministrator(c)) {
@@ -528,7 +539,7 @@ MCreadClient(Client c)
 	bstream *in = c->fdin;
 
 #ifdef MAL_CLIENT_DEBUG
-	printf("# streamClient %d %d\n", c->idx, isa_block_stream(in->s));
+	fprintf(stderr,"# streamClient %d %d\n", c->idx, isa_block_stream(in->s));
 #endif
 
 	while (in->pos < in->len &&
@@ -563,13 +574,13 @@ MCreadClient(Client c)
 				in->len++;
 		}
 #ifdef MAL_CLIENT_DEBUG
-		printf("# simple stream received %d sum " SZFMT "\n", c->idx, sum);
+		fprintf(stderr, "# simple stream received %d sum " SZFMT "\n", c->idx, sum);
 #endif
 	}
 	if (in->pos >= in->len) {
 		/* end of stream reached */
 #ifdef MAL_CLIENT_DEBUG
-		printf("# end of stream received %d %d\n", c->idx, c->bak == 0);
+		fprintf(stderr,"# end of stream received %d %d\n", c->idx, c->bak == 0);
 #endif
 		if (c->bak) {
 			MCpopClientInput(c);
@@ -579,13 +590,8 @@ MCreadClient(Client c)
 		}
 		return 0;
 	}
-	if (*CURRENT(c) == '?') {
-		showHelp(c->nspace, CURRENT(c) + 1, c->fdout);
-		in->pos = in->len;
-		return MCreadClient(c);
-	}
 #ifdef MAL_CLIENT_DEBUG
-	printf("# finished stream read %d %d\n", (int) in->pos, (int) in->len);
+	fprintf(stderr,"# finished stream read %d %d\n", (int) in->pos, (int) in->len);
 	printf("#%s\n", in->buf);
 #endif
 	return 1;
