@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef _MAL_FCN_H
@@ -13,13 +13,9 @@
 #include "mal_module.h"
 #include "mal_resolve.h"
 
-typedef struct lifespan {
-	int beginLifespan, endLifespan, lastUpdate;
-} *Lifespan, LifespanRecord;
-
-#define getLastUpdate(L,I)	(L[I].lastUpdate)
-#define getEndLifespan(L,I)	(L[I].endLifespan)
-#define getBeginLifespan(L,I)	(L[I].beginLifespan)
+#define getLastUpdate(L,I)	((L)->var[I].updated)
+#define getEndScope(L,I)	((L)->var[I].eolife)
+#define getBeginScope(L,I)	((L)->var[I].declared)
 
 /* #define DEBUG_MAL_FCN */
 /* #define DEBUG_CLONE */
@@ -34,13 +30,11 @@ mal_export void clrDeclarations(MalBlkPtr mb);
 mal_export int isLoopBarrier(MalBlkPtr mb, int pc);
 mal_export int getBlockExit(MalBlkPtr mb,int pc);
 mal_export int getBlockBegin(MalBlkPtr mb,int pc);
-
-#define newLifespan(M) (Lifespan)GDKzalloc(sizeof(LifespanRecord)*(M)->vsize)
-mal_export Lifespan setLifespan(MalBlkPtr mb);
+mal_export void setVariableScope(MalBlkPtr mb);
 
 mal_export void printFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg);
+mal_export void fprintFunction(FILE *fd, MalBlkPtr mb, MalStkPtr stk, int flg);
 mal_export void debugFunction(stream *fd, MalBlkPtr mb, MalStkPtr stk, int flg, int first, int size);
-mal_export void showFlowGraph(MalBlkPtr mb, MalStkPtr stk, str fname);
 
 #include "mal_exception.h"
 
