@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -35,7 +35,6 @@ ptr getArgReference(MalStkPtr stk, InstrPtr pci, int k)
 	case TYPE_sht:  return (ptr) &v->val.shval;
 	case TYPE_bat:  return (ptr) &v->val.bval;
 	case TYPE_int:  return (ptr) &v->val.ival;
-	case TYPE_wrd:  return (ptr) &v->val.wval;
 	case TYPE_bte:  return (ptr) &v->val.btval;
 	case TYPE_oid:  return (ptr) &v->val.oval;
 	case TYPE_ptr:  return (ptr) &v->val.pval;
@@ -70,34 +69,34 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 	str ret= MAL_SUCCEED;
 
 	switch (pci->argc) {
-	case 0: ret = (str)(*pci->fcn)();
+	case 0: ret = (*pci->fcn)();
 		break;
-	case 1: ret = (str)(*pci->fcn)(
+	case 1: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0));
 		break;
-	case 2: ret = (str)(*pci->fcn)(
+	case 2: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1));
 		break;
-	case 3: ret = (str)(*pci->fcn)(
+	case 3: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2));
 		break;
-	case 4: ret = (str)(*pci->fcn)(
+	case 4: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
 			getArgReference(stk, pci, 3));
 		break;
-	case 5: ret = (str)(*pci->fcn)(
+	case 5: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
 			getArgReference(stk, pci, 3),
 			getArgReference(stk, pci, 4));
 		break;
-	case 6: ret = (str)(*pci->fcn)(
+	case 6: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -105,7 +104,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 4),
 			getArgReference(stk, pci, 5));
 		break;
-	case 7: ret = (str)(*pci->fcn)(
+	case 7: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -114,7 +113,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 5),
 			getArgReference(stk, pci, 6));
 		break;
-	case 8: ret = (str)(*pci->fcn)(
+	case 8: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -124,7 +123,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 6),
 			getArgReference(stk, pci, 7));
 		break;
-	case 9: ret = (str)(*pci->fcn)(
+	case 9: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -135,7 +134,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 7),
 			getArgReference(stk, pci, 8));
 		break;
-	case 10: ret = (str)(*pci->fcn)(
+	case 10: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -147,7 +146,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 8),
 			getArgReference(stk, pci, 9));
 		break;
-	case 11: ret = (str)(*pci->fcn)(
+	case 11: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -160,7 +159,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 9),
 			getArgReference(stk, pci, 10));
 		break;
-	case 12: ret = (str)(*pci->fcn)(
+	case 12: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -174,7 +173,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 10),
 			getArgReference(stk, pci, 11));
 		break;
-	case 13: ret = (str)(*pci->fcn)(
+	case 13: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -189,7 +188,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 11),
 			getArgReference(stk, pci, 12));
 		break;
-	case 14: ret = (str)(*pci->fcn)(
+	case 14: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -205,7 +204,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 12),
 			getArgReference(stk, pci, 13));
 		break;
-	case 15: ret = (str)(*pci->fcn)(
+	case 15: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -222,7 +221,7 @@ str malCommandCall(MalStkPtr stk, InstrPtr pci)
 			getArgReference(stk, pci, 13),
 			getArgReference(stk, pci, 14));
 		break;
-	case 16: ret = (str)(*pci->fcn)(
+	case 16: ret = (*pci->fcn)(
 			getArgReference(stk, pci, 0),
 			getArgReference(stk, pci, 1),
 			getArgReference(stk, pci, 2),
@@ -283,8 +282,10 @@ prepareMALstack(MalBlkPtr mb, int size)
 	int i;
 	ValPtr lhs, rhs;
 
-	assert(size >= mb->vsize);
 	stk = newGlobalStack(size);
+	if (!stk) {
+		return NULL;
+	}
 	//memset((char *)stk, 0, stackSize(size)); already set
 	//stk->stksize = size;
 	stk->stktop = mb->vtop;
@@ -320,9 +321,9 @@ str runMAL(Client cntxt, MalBlkPtr mb, MalBlkPtr mbcaller, MalStkPtr env)
 	if (env != NULL) {
 		stk = env;
 		if (mb != stk->blk)
-			showScriptException(cntxt->fdout, mb, 0, MAL, "runMAL:misalignment of symbols\n");
+			throw(MAL, "mal.interpreter","misalignment of symbols");
 		if (mb->vtop > stk->stksize)
-			showScriptException(cntxt->fdout, mb, 0, MAL, "stack too small\n");
+			throw(MAL, "mal.interpreter","stack too small");
 		initStack(env->stkbot);
 	} else {
 		stk = prepareMALstack(mb, mb->vsize);
@@ -357,7 +358,7 @@ str runMAL(Client cntxt, MalBlkPtr mb, MalBlkPtr mbcaller, MalStkPtr env)
 	if (!stk->keepAlive && garbageControl(getInstrPtr(mb, 0)))
 		garbageCollector(cntxt, mb, stk, env != stk);
 	if (stk && stk != env)
-		GDKfree(stk);
+		freeStack(stk);
 	if (cntxt->qtimeout && GDKusec()- mb->starttime > cntxt->qtimeout)
 		throw(MAL, "mal.interpreter", RUNTIME_QRY_TIMEOUT);
 	return ret;
@@ -407,8 +408,8 @@ callMAL(Client cntxt, MalBlkPtr mb, MalStkPtr *env, ValPtr argv[], char debug)
 
 	cntxt->lastcmd= time(0);
 #ifdef DEBUG_CALLMAL
-	mnstr_printf(cntxt->fdout, "callMAL\n");
-	printInstruction(cntxt->fdout, mb, 0, pci, LIST_MAL_ALL);
+	fprintf(stderr, "callMAL\n");
+	fprintInstruction(stderr, mb, 0, pci, LIST_MAL_ALL);
 #endif
 	switch (pci->token) {
 	case FUNCTIONsymbol:
@@ -422,13 +423,19 @@ callMAL(Client cntxt, MalBlkPtr mb, MalStkPtr *env, ValPtr argv[], char debug)
 			stk = prepareMALstack(mb, mb->vsize);
 			stk->up = 0;
 			*env = stk;
-		} else stk = *env;
+		} else {
+			ValPtr lhs, rhs;
+
+			stk = *env;
+			initStack(0);
+		}
 		assert(stk);
 		for (i = pci->retc; i < pci->argc; i++) {
 			lhs = &stk->stk[pci->argv[i]];
-			VALcopy(lhs, argv[i]);
+			if (VALcopy(lhs, argv[i]) == NULL)
+				throw(MAL, "mal.interpreter", MAL_MALLOC_FAIL);
 			if (lhs->vtype == TYPE_bat)
-				BBPincref(lhs->val.bval, TRUE);
+				BBPretain(lhs->val.bval);
 		}
 		stk->cmd = debug;
 		ret = runMALsequence(cntxt, mb, 1, 0, stk, 0, 0);
@@ -444,6 +451,8 @@ callMAL(Client cntxt, MalBlkPtr mb, MalStkPtr *env, ValPtr argv[], char debug)
 	}
 	if ( ret == MAL_SUCCEED && cntxt->qtimeout && GDKusec()- mb->starttime > cntxt->qtimeout)
 		throw(MAL, "mal.interpreter", RUNTIME_QRY_TIMEOUT);
+	if (stk) 
+		garbageCollector(cntxt, mb, stk, TRUE);
 	return ret;
 }
 
@@ -462,13 +471,13 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 	InstrPtr pci = 0;
 	int exceptionVar;
 	str ret = 0, localGDKerrbuf= GDKerrbuf;
-	int stamp = -1;
 	ValRecord backups[16];
 	ValPtr backup;
 	int garbages[16], *garbage;
 	int stkpc = 0;
 	RuntimeProfileRecord runtimeProfile, runtimeProfileFunction;
 	lng lastcheck = 0;
+	int	startedProfileQueue = 0;
 #define CHECKINTERVAL 1000 /* how often do we check for client disconnect */
 	runtimeProfile.ticks = runtimeProfileFunction.ticks = 0;
 
@@ -479,24 +488,37 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 	if (startpc+1 == stoppc) {
 		pci = getInstrPtr(mb, startpc);
 		if (pci->argc > 16) {
-			backup = GDKzalloc(pci->argc * sizeof(ValRecord));
+			backup = GDKmalloc(pci->argc * sizeof(ValRecord));
+			if( backup == NULL)
+				throw(MAL, "mal.interpreter", MAL_MALLOC_FAIL);
 			garbage = (int*)GDKzalloc(pci->argc * sizeof(int));
+			if( garbage == NULL){
+				GDKfree(backup);
+				throw(MAL, "mal.interpreter", MAL_MALLOC_FAIL);
+			}
 		} else {
 			backup = backups;
 			garbage = garbages;
-			memset((char*) garbages, 0, 16 * sizeof(int));
+			memset(garbages, 0, sizeof(garbages));
 		}
 	} else if ( mb->maxarg > 16 ){
-		backup = GDKzalloc(mb->maxarg * sizeof(ValRecord));
+		backup = GDKmalloc(mb->maxarg * sizeof(ValRecord));
+		if( backup == NULL)
+			throw(MAL, "mal.interpreter", MAL_MALLOC_FAIL);
 		garbage = (int*)GDKzalloc(mb->maxarg * sizeof(int));
+		if( garbage == NULL){
+			GDKfree(backup);
+			throw(MAL, "mal.interpreter", MAL_MALLOC_FAIL);
+		}
 	} else {
 		backup = backups;
 		garbage = garbages;
-		memset((char*) garbages, 0, 16 * sizeof(int));
+		memset(garbages, 0, sizeof(garbages));
 	}
 
 	/* also produce event record for start of function */
 	if ( startpc == 1 &&  startpc < mb->stop ){
+		startedProfileQueue = 1;
 		runtimeProfileInit(cntxt, mb, stk);
 		runtimeProfileBegin(cntxt, mb, stk, getInstrPtr(mb,0), &runtimeProfileFunction);
 		mb->starttime = GDKusec();
@@ -529,6 +551,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 
 			if (stk->cmd == 0)
 				stk->cmd = cntxt->itrace;
+			//mdbStep(cntxt, mb, stk, stkpc);
 			if (stk->cmd == 'x' ) {
 				stk->cmd = 0;
 				stkpc = mb->stop;
@@ -548,9 +571,10 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			lastcheck = runtimeProfile.ticks;
 		}
 
+
 		/* The interpreter loop
 		 * The interpreter is geared towards execution a MAL
-		 * procedure together with all its decendant
+		 * procedure together with all its descendant
 		 * invocations. As such, it provides the MAL abtract
 		 * machine processor.
 		 *
@@ -568,30 +592,19 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 		 * garbage collected are identified. In the post-execution
 		 * phase they are removed.
 		 */
+		for (i = 0; i < pci->retc; i++)
+			backup[i] = stk->stk[getArg(pci, i)];
+
 		if (garbageControl(pci)) {
 			for (i = 0; i < pci->argc; i++) {
 				int a = getArg(pci, i);
-
-				backup[i].vtype = 0;
-				backup[i].len = 0;
-				backup[i].val.pval = 0;
 				garbage[i] = -1;
-				if (stk->stk[a].vtype == TYPE_bat && getEndOfLife(mb, a) == stkpc && isNotUsedIn(pci, i + 1, a))
+				if (stk->stk[a].vtype == TYPE_bat && getEndScope(mb, a) == stkpc && isNotUsedIn(pci, i + 1, a))
 					garbage[i] = a;
-
-				if (i < pci->retc && stk->stk[a].vtype == TYPE_bat) {
-					backup[i] = stk->stk[a];
-					stamp = BBPcurstamp();
-				} else if (i < pci->retc &&
-						   0 < stk->stk[a].vtype &&
-						   stk->stk[a].vtype < TYPE_any &&
-						   ATOMextern(stk->stk[a].vtype)) {
-					backup[i] = stk->stk[a];
-				}
 			}
 		}
 
-		FREE_EXCEPTION(ret);
+		freeException(ret);
 		ret = 0;
 		switch (pci->token) {
 		case ASSIGNsymbol:
@@ -615,9 +628,9 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				rhs = &stk->stk[pci->argv[i]];
 				VALcopy(lhs, rhs);
 				if (lhs->vtype == TYPE_bat && lhs->val.bval != bat_nil)
-					BBPincref(lhs->val.bval, TRUE);
+					BBPretain(lhs->val.bval);
 			}
-			FREE_EXCEPTION(ret);
+			freeException(ret);
 			ret = 0;
 			break;
 		case PATcall:
@@ -625,7 +638,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				ret = createScriptException(mb, stkpc, MAL, NULL,
 					"address of pattern %s.%s missing", pci->modname, pci->fcnname);
 			} else {
-				ret = (str)(*pci->fcn)(cntxt, mb, stk, pci);
+				ret = (*pci->fcn)(cntxt, mb, stk, pci);
 #ifndef NDEBUG
 				/* check that the types of actual results match
 				 * expected results */
@@ -636,7 +649,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					if (isaBatType(t)) {
 						bat bid = stk->stk[a].val.bval;
 						BAT *_b = BATdescriptor(bid);
-						t = getColumnType(t);
+						t = getBatType(t);
 						assert(stk->stk[a].vtype == TYPE_bat);
 						assert(bid == 0 ||
 							   bid == bat_nil ||
@@ -661,12 +674,12 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 
 				if (isaBatType(t)) {
 					bat bid = stk->stk[a].val.bval;
-					t = getColumnType(t);
+					t = getBatType(t);
 					assert(stk->stk[a].vtype == TYPE_bat);
 					assert(bid == 0 ||
 						   bid == bat_nil ||
 						   t == TYPE_any ||
-						   ATOMtype(BBP_cache(bid)->ttype) == ATOMtype(t));
+						   ATOMtype(BBP_desc(bid)->ttype) == ATOMtype(t));
 				} else {
 					assert(t == stk->stk[a].vtype);
 				}
@@ -720,11 +733,13 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				nstk->up = stk;
 				if (nstk->calldepth > 256) {
 					ret= createException(MAL, "mal.interpreter", MAL_CALLDEPTH_FAIL);
+					GDKfree(nstk);
 					break;
 				}
 				if ((unsigned)nstk->stkdepth > THREAD_STACK_SIZE / sizeof(mb->var[0]) / 4 && THRhighwater()){
 					/* we are running low on stack space */
 					ret= createException(MAL, "mal.interpreter", MAL_STACK_FAIL);
+					GDKfree(nstk);
 					break;
 				}
 
@@ -736,9 +751,12 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					rhs = &stk->stk[pci->argv[ii]];
 					VALcopy(lhs, rhs);
 					if (lhs->vtype == TYPE_bat)
-						BBPincref(lhs->val.bval, TRUE);
+						BBPretain(lhs->val.bval);
 				}
 				ret = runMALsequence(cntxt, pci->blk, 1, pci->blk->stop, nstk, stk, pci);
+				for (ii = 0; ii < nstk->stktop; ii++)
+					if (ATOMextern(nstk->stk[ii].vtype))
+						GDKfree(nstk->stk[ii].val.pval);
 				GDKfree(nstk);
 			}
 			break;
@@ -750,7 +768,6 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				ret = shutdownFactory(cntxt, mb);
 			runtimeProfileExit(cntxt, mb, stk, pci, &runtimeProfile);
 			runtimeProfileExit(cntxt, mb, stk, getInstrPtr(mb,0), &runtimeProfileFunction);
-			runtimeProfileFinish(cntxt, mb);
 			if (pcicaller && garbageControl(getInstrPtr(mb, 0)))
 				garbageCollector(cntxt, mb, stk, TRUE);
 			if (cntxt->qtimeout && GDKusec()- mb->starttime > cntxt->qtimeout){
@@ -783,10 +800,18 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 		if( mb->stop <= 1)
 			continue;
 		runtimeProfileExit(cntxt, mb, stk, pci, &runtimeProfile);
-		if (ret != MAL_SUCCEED)
-			runtimeProfileFinish(cntxt, mb);
 		/* check for strong debugging after each MAL statement */
 		if ( pci->token != FACcall && ret== MAL_SUCCEED) {
+			for (i = 0; i < pci->retc; i++) {
+				lhs = &backup[i];
+				if (BATatoms[lhs->vtype].atomUnfix)
+					(*BATatoms[lhs->vtype].atomUnfix)(VALget(lhs));
+				if (ATOMextern(lhs->vtype) &&
+					lhs->val.pval &&
+					lhs->val.pval != ATOMnilptr(lhs->vtype) &&
+					lhs->val.pval != stk->stk[getArg(pci, i)].val.pval)
+					GDKfree(lhs->val.pval);
+			}
 			if (GDKdebug & (CHECKMASK|PROPMASK) && exceptionVar < 0) {
 				BAT *b;
 
@@ -794,23 +819,16 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					if (garbage[i] == -1 && stk->stk[getArg(pci, i)].vtype == TYPE_bat &&
 						stk->stk[getArg(pci, i)].val.bval != bat_nil &&
 						stk->stk[getArg(pci, i)].val.bval != 0) {
-						b = BBPquickdesc(abs(stk->stk[getArg(pci, i)].val.bval), FALSE);
+						assert(stk->stk[getArg(pci, i)].val.bval > 0);
+						b = BBPquickdesc(stk->stk[getArg(pci, i)].val.bval, FALSE);
 						if (b == NULL) {
 							if (ret == MAL_SUCCEED)
 								ret = createException(MAL, "mal.propertyCheck", RUNTIME_OBJECT_MISSING);
 							continue;
 						}
-						if (b->batStamp <= stamp) {
-							if (GDKdebug & PROPMASK) {
-								b = BATdescriptor(stk->stk[getArg(pci, i)].val.bval);
-								BATassertProps(b);
-								BBPunfix(b->batCacheid);
-							}
-						} else if (GDKdebug & CHECKMASK) {
-							b = BATdescriptor(stk->stk[getArg(pci, i)].val.bval);
-							BATassertProps(b);
-							BBPunfix(b->batCacheid);
-						}
+						b = BATdescriptor(stk->stk[getArg(pci, i)].val.bval);
+						BATassertProps(b);
+						BBPunfix(b->batCacheid);
 					}
 				}
 			}
@@ -824,31 +842,11 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					if (isaBatType(getArgType(mb, pci, i))) {
 						bat bid = stk->stk[a].val.bval;
 
-						if (i < pci->retc && backup[i].val.bval != bat_nil) {
-							bat bx = backup[i].val.bval;
-							backup[i].val.bval = bat_nil;
-							BBPdecref(bx, TRUE);
-						}
 						if (garbage[i] >= 0) {
-							PARDEBUG mnstr_printf(GDKstdout, "#GC pc=%d bid=%d %s done\n", stkpc, bid, getVarName(mb, garbage[i]));
-							bid = abs(stk->stk[garbage[i]].val.bval);
+							PARDEBUG fprintf(stderr, "#GC pc=%d bid=%d %s done\n", stkpc, bid, getVarName(mb, garbage[i]));
+							bid = stk->stk[garbage[i]].val.bval;
 							stk->stk[garbage[i]].val.bval = bat_nil;
-							BBPdecref(bid, TRUE);
-						}
-					} else if (i < pci->retc &&
-							   0 < stk->stk[a].vtype &&
-							   stk->stk[a].vtype < TYPE_any &&
-							   ATOMextern(stk->stk[a].vtype)) {
-						if (backup[i].val.pval &&
-							backup[i].val.pval != stk->stk[a].val.pval) {
-							if (backup[i].val.pval)
-								GDKfree(backup[i].val.pval);
-							if (i >= pci->retc) {
-								stk->stk[a].val.pval = 0;
-								stk->stk[a].len = 0;
-							}
-							backup[i].len = 0;
-							backup[i].val.pval = 0;
+							BBPrelease(bid);
 						}
 					}
 				}
@@ -860,12 +858,13 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			str oldret = ret;
 			ret = catchKernelException(cntxt, ret);
 			if (ret != oldret)
-				FREE_EXCEPTION(oldret);
+				freeException(oldret);
 		}
 
 		if (ret != MAL_SUCCEED) {
 			str msg = 0;
 
+
 			if (stk->cmd || mb->trap) {
 				mnstr_printf(cntxt->fdout, "!ERROR: %s\n", ret);
 				stk->cmd = '\n'; /* in debugging go to step mode */
@@ -883,7 +882,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			/* Detect any exception received from the implementation. */
 			/* The first identifier is an optional exception name */
 			if (strstr(ret, "!skip-to-end")) {
-				GDKfree(ret);       /* no need to check for M5OutOfMemory */
+				freeException(ret);
 				ret = MAL_SUCCEED;
 				stkpc = mb->stop;
 				continue;
@@ -900,7 +899,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				*msg = ':';
 			}
 			if (exceptionVar == -1)
-				exceptionVar = findVariableLength(mb, (str)"ANYexception", 12);
+				exceptionVar = findVariableLength(mb, "ANYexception", 12);
 
 			/* unknown exceptions lead to propagation */
 			if (exceptionVar == -1) {
@@ -915,7 +914,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				MT_lock_set(&mal_contextLock);
 				v = &stk->stk[exceptionVar];
 				if (v->val.sval)
-					FREE_EXCEPTION(v->val.sval);    /* old exception*/
+					freeException(v->val.sval);    /* old exception*/
 				v->vtype = TYPE_str;
 				v->val.sval = ret;
 				v->len = (int)strlen(v->val.sval);
@@ -923,16 +922,18 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				MT_lock_unset(&mal_contextLock);
 			} else {
 				mnstr_printf(cntxt->fdout, "%s", ret);
-				FREE_EXCEPTION(ret);
+				freeException(ret);
 			}
 			/* position yourself at the catch instruction for further decisions */
 			/* skipToCatch(exceptionVar,@2,@3) */
 			if (stk->cmd == 'C' || mb->trap) {
 				stk->cmd = 'n';
+				//mdbStep(cntxt, mb, stk, stkpc);
 				if (stk->cmd == 'x' ) {
 					stkpc = mb->stop;
 					continue;
 				}
+
 			}
 			/* skip to catch block or end */
 			for (; stkpc < mb->stop; stkpc++) {
@@ -956,10 +957,10 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				}
 				continue;
 			}
+
 			pci = getInstrPtr(mb, stkpc);
 		}
 
-
 		/*
 		 * After the expression has been evaluated we should check for
 		 * a possible change in the control flow.
@@ -1053,12 +1054,6 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 				else
 					stkpc++;
 				break;
-			case TYPE_wrd:
-				if (v->val.wval != wrd_nil)
-					stkpc = pci->jump;
-				else
-					stkpc++;
-				break;
 			case TYPE_bte:
 				if (v->val.btval != bte_nil)
 					stkpc = pci->jump;
@@ -1112,7 +1107,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			break;
 		case RAISEsymbol:
 			exceptionVar = getDestVar(pci);
-			FREE_EXCEPTION(ret);
+			freeException(ret);
 			ret = NULL;
 			if (getVarType(mb, getDestVar(pci)) == TYPE_str) {
 				ret = createScriptException(mb, stkpc, MAL, NULL,
@@ -1121,6 +1116,7 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			/* skipToCatch(exceptionVar, @2, stk) */
 			if (stk->cmd == 'C' || mb->trap) {
 				stk->cmd = 'n';
+				//mdbStep(cntxt, mb, stk, stkpc);
 				if (stk->cmd == 'x' ) {
 					stkpc = mb->stop;
 					continue;
@@ -1144,7 +1140,6 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 			if (stkpc == mb->stop) {
 				runtimeProfileExit(cntxt, mb, stk, pci, &runtimeProfile);
 				runtimeProfileExit(cntxt, mb, stk, getInstrPtr(mb,0), &runtimeProfileFunction);
-				runtimeProfileFinish(cntxt, mb);
 				break;
 			}
 			if (stkpc == mb->stop)
@@ -1152,6 +1147,8 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					"Exception raised");
 			break;
 		case YIELDsymbol:     /* to be defined */
+			if( startedProfileQueue)
+				runtimeProfileFinish(cntxt, mb, stk);
 			if ( backup != backups) GDKfree(backup);
 			if ( garbage != garbages) GDKfree(garbage);
 			return yieldFactory(mb, pci, stkpc);
@@ -1171,14 +1168,13 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 						lhs = &env->stk[pci->argv[i]];
 						VALcopy(lhs, rhs);
 						if (lhs->vtype == TYPE_bat)
-							BBPincref(lhs->val.bval, TRUE);
+							BBPretain(lhs->val.bval);
 					}
 					if (garbageControl(getInstrPtr(mb, 0)))
 						garbageCollector(cntxt, mb, stk, TRUE);
 					/* reset the clock */
 					runtimeProfileExit(cntxt, mb, stk, pp, &runtimeProfile);
 					runtimeProfileExit(cntxt, mb, stk, getInstrPtr(mb,0), &runtimeProfileFunction);
-					runtimeProfileFinish(cntxt, mb);
 				} 
 			}
 			stkpc = mb->stop;
@@ -1210,8 +1206,10 @@ str runMALsequence(Client cntxt, MalBlkPtr mb, int startpc,
 					NULL, "Exception not caught");
 			}
 		}
-		FREE_EXCEPTION(oldret);
+		freeException(oldret);
 	}
+	if( startedProfileQueue)
+		runtimeProfileFinish(cntxt, mb, stk);
 	if ( backup != backups) GDKfree(backup);
 	if ( garbage != garbages) GDKfree(garbage);
 	return ret;
@@ -1327,12 +1325,15 @@ str catchKernelException(Client cntxt, str ret)
 				strcpy(z, ret);
 				if (z[strlen(z) - 1] != '\n') strcat(z, "\n");
 				strcat(z, errbuf);
-			}
+			} else // when malloc fails, leave it somewhere
+				fprintf(stderr,"!catchKernelException:%s\n",ret);
 		} else {
 			/* trap hidden (GDK) exception */
 			z = (char*)GDKmalloc(strlen("GDKerror:") + strlen(errbuf) + 2);
 			if (z)
 				sprintf(z, "GDKerror:%s", errbuf);
+			else
+				fprintf(stderr,"!catchKernelException:GDKerror:%s\n",errbuf);
 		}
 		/* did we eat the error away of not */
 		if (z)
@@ -1378,7 +1379,7 @@ void garbageElement(Client cntxt, ValPtr v)
 		 * allowed during the execution of a GDK operation.
 		 * All references should be logical.
 		 */
-		bat bid = abs(v->val.bval);
+		bat bid = v->val.bval;
 		/* printf("garbage collecting: %d lrefs=%d refs=%d\n",
 		   bid, BBP_lrefs(bid),BBP_refs(bid));*/
 		v->val.bval = bat_nil;
@@ -1386,7 +1387,7 @@ void garbageElement(Client cntxt, ValPtr v)
 			return;
 		if (!BBP_lrefs(bid))
 			return;
-		BBPdecref(bid, TRUE);
+		BBPrelease(bid);
 	} else if (0 < v->vtype && v->vtype < MAXATOMS && ATOMextern(v->vtype)) {
 		if (v->val.pval)
 			GDKfree(v->val.pval);
@@ -1415,13 +1416,27 @@ void garbageCollector(Client cntxt, MalBlkPtr mb, MalStkPtr stk, int flag)
 {
 	int k;
 	ValPtr v;
-	(void)cntxt;
 
+#ifdef STACKTRACE
+	if (cntxt) {
+		mnstr_printf(cntxt->fdout, "#--->stack before garbage collector\n");
+		printStack(cntxt->fdout, mb, stk, 0);
+	}
+#endif
+	(void) flag;
 	for (k = 0; k < mb->vtop; k++) {
-		if (isVarCleanup(mb, k) && (flag || isTmpVar(mb, k))) {
+	//	if (isVarCleanup(mb, k) ){
 			garbageElement(cntxt, v = &stk->stk[k]);
 			v->vtype = TYPE_int;
 			v->val.ival = int_nil;
-		}
+	//	}
 	}
+#ifdef STACKTRACE
+	if (cntxt) {
+		mnstr_printf(cntxt->fdout, "#-->stack after garbage collector\n");
+		printStack(cntxt->fdout, mb, stk, 0);
+	}
+#else
+	(void)cntxt;
+#endif
 }
