@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef _OPT_COERCION_
@@ -12,8 +12,6 @@
 #include "mal_interpreter.h"
 #include "opt_support.h"
 
-opt_export int OPTcoercionImplementation(Client cntxt,MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
-
-#define OPTDEBUGcoercion  if ( optDebug & ((lng)1 << DEBUG_OPT_COERCION) )
+mal_export str OPTcoercionImplementation(Client cntxt,MalBlkPtr mb, MalStkPtr stk, InstrPtr pci);
 
 #endif
