@@ -1,34 +1,41 @@
+#include "monetdb_config.h"
+
+#ifdef HAVE_EMBEDDED_R
 #include <R.h>
 #include <Rdefines.h>
 #include "mapisplit.h"
+#include "mapisplit-r.h"
 
 char nullstr[] = "NULL";
 
 SEXP mapi_split(SEXP mapiLinesVector, SEXP numCols) {
-	int cols = INTEGER_POINTER(AS_INTEGER(numCols))[0];
-	int rows = LENGTH(mapiLinesVector);
+	const int cols = INTEGER_POINTER(AS_INTEGER(numCols))[0];
+	const int rows = LENGTH(mapiLinesVector);
+	SEXP colVec;
+	int cRow;
+	int cCol;
+	char **elems = malloc(sizeof(char*)* cols);
+	if (!elems) {
+		error("Memory allocation failure");
+	}
 
 	if (!IS_CHARACTER(mapiLinesVector) || rows < 1 || cols < 1) {
 		error("Invalid input to mapi_split: type=%d, rows=%d, cols=%d", TYPEOF(mapiLinesVector), rows, cols);
 	}
 
-	SEXP colVec;
 	PROTECT(colVec = NEW_LIST(cols));
 
-	int col;
-	for (col = 0; col < cols; col++) {
+	for (cRow = 0; cRow < cols; cRow++) {
 		SEXP colV = PROTECT(NEW_STRING(rows));
-		SET_ELEMENT(colVec, col, colV);
+		SET_ELEMENT(colVec, cRow, colV);
 		UNPROTECT(1);
 	}
 
-	int cRow;
-	int cCol;
-	char* elems[cols];
 
 	for (cRow = 0; cRow < rows; cRow++) {
 		const char *rval = CHAR(STRING_ELT(mapiLinesVector, cRow));
 		char *val = strdup(rval);
+
 		cCol = 0;
 		mapi_line_split(val, elems, cols);
 
@@ -44,7 +51,9 @@ SEXP mapi_split(SEXP mapiLinesVector, SEXP numCols) {
 		}
 		free(val);
 	}
+	free(elems);
 
 	UNPROTECT(1);
 	return colVec;
 }
+#endif
