@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #include "monetdb_config.h"
@@ -20,7 +20,6 @@
 #include <rel_exp.h>
 #include <rel_dump.h>
 #include <rel_bin.h>
-#include "clients.h"
 #include "mal_instruction.h"
 
 #define CONCAT_2(a, b)		a##b
@@ -48,10 +47,6 @@
 #include "sql_round_impl.h"
 #undef TYPE
 
-#define TYPE wrd
-#include "sql_round_impl.h"
-#undef TYPE
-
 #define TYPE lng
 #include "sql_round_impl.h"
 #undef TYPE
