@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -35,6 +35,7 @@ list *types = NULL;
 list *aggrs = NULL;
 list *funcs = NULL;
 
+static sql_type *BIT = NULL;
 static list *localtypes = NULL;
 
 int digits2bits(int digits) 
@@ -43,6 +44,12 @@ int digits2bits(int digits)
 		return 8;
 	else if (digits < 5) 
 		return 16;
+	else if (digits <= 5) 
+		return 17;
+	else if (digits <= 6) 
+		return 20;
+	else if (digits <= 7) 
+		return 24;
 	else if (digits <= 8) 
 		return 27;
 	else if (digits < 10) 
@@ -168,11 +175,11 @@ localtypes_cmp(int nlt, int olt)
 	if (nlt == TYPE_flt || nlt == TYPE_dbl) {
 		nlt = TYPE_dbl;
 #ifdef HAVE_HGE
-	} else if (nlt == TYPE_bte || nlt == TYPE_sht || nlt == TYPE_int || nlt == TYPE_wrd || nlt == TYPE_lng || nlt == TYPE_hge) {
+	} else if (nlt == TYPE_bte || nlt == TYPE_sht || nlt == TYPE_int || nlt == TYPE_lng || nlt == TYPE_hge) {
 		assert(have_hge || nlt != TYPE_hge);
 		nlt = have_hge ? TYPE_hge : TYPE_lng;
 #else
-	} else if (nlt == TYPE_bte || nlt == TYPE_sht || nlt == TYPE_int || nlt == TYPE_wrd || nlt == TYPE_lng) {
+	} else if (nlt == TYPE_bte || nlt == TYPE_sht || nlt == TYPE_int || nlt == TYPE_lng) {
 		nlt = TYPE_lng;
 #endif
 	}
@@ -192,14 +199,14 @@ sql_find_numeric(sql_subtype *r, int localtype, unsigned int digits)
 #ifdef HAVE_HGE
 		if (have_hge) {
 			localtype = TYPE_hge;
-			if (digits > 128)
-				digits = 128;
+			if (digits >= 128)
+				digits = 127;
 		} else
 #endif
 		{
 			localtype = TYPE_lng;
-			if (digits > 64)
-				digits = 64;
+			if (digits >= 64)
+				digits = 63;
 		}
 	}
 
@@ -207,7 +214,7 @@ sql_find_numeric(sql_subtype *r, int localtype, unsigned int digits)
 		sql_type *t = n->data;
 
 		if (localtypes_cmp(t->localtype, localtype)) {
-			if ((digits && t->digits >= digits) || (digits == t->digits)) {
+			if ((digits && t->digits > digits) || (!digits && digits == t->digits)) {
 				sql_init_subtype(r, t, digits, 0);
 				return r;
 			}
@@ -217,7 +224,7 @@ sql_find_numeric(sql_subtype *r, int localtype, unsigned int digits)
 					break;
 				}
 				n = m;
-				if ((digits && t->digits >= digits) || (digits == t->digits)) {
+				if ((digits && t->digits > digits) || (!digits && digits == t->digits)) {
 					sql_init_subtype(r, t, digits, 0);
 					return r;
 				}
@@ -630,7 +637,10 @@ sql_dup_subfunc(sql_allocator *sa, sql_func *f, list *ops, sql_subtype *member)
 	sql_subfunc *fres = SA_ZNEW(sa, sql_subfunc);
 
 	fres->func = f;
-	if (IS_FUNC(f) || IS_UNION(f) || IS_ANALYTIC(f)) { /* not needed for PROC/FILT */
+	if (IS_FILT(f)) {
+		fres->res = sa_list(sa);
+		list_append(fres->res, sql_bind_localtype("bit"));
+	} else if (IS_FUNC(f) || IS_UNION(f) || IS_ANALYTIC(f)) { /* not needed for PROC */
 		unsigned int mscale = 0, mdigits = 0;
 
 		if (ops) for (tn = ops->h; tn; tn = tn->next) {
@@ -681,6 +691,21 @@ sql_dup_subfunc(sql_allocator *sa, sql_func *f, list *ops, sql_subtype *member)
 				list_append(fres->res, res);
 			}
 		}
+		if (member) { /* check that the types of all EC_ANY's are equal */
+			sql_subtype *st = NULL;
+			node *m;
+
+			if (ops) for (tn = ops->h, m = f->ops->h; tn; tn = tn->next, m = m->next) {
+				sql_arg *s = m->data;
+
+				if (s->type.type->eclass == EC_ANY) {
+					if (!st)
+						st = tn->data;
+					else if (subtype_cmp(st, tn->data))
+						return NULL;
+				}
+			}
+		}
 	}
 	return fres;
 }
@@ -704,6 +729,7 @@ sql_find_func(sql_allocator *sa, sql_schema *s, const char *sqlfname, int nrargs
 	int key = hash_key(sqlfname);
 	sql_hash_e *he;
 	int found = 0;
+	int filt = (type == F_FUNC)?F_FILT:type;
 
 	assert(nrargs);
 	MT_lock_set(&funcs->ht_lock);
@@ -718,7 +744,7 @@ sql_find_func(sql_allocator *sa, sql_schema *s, const char *sqlfname, int nrargs
 	for (; he; he = he->chain) {
 		sql_func *f = he->value;
 
-		if (f->type != type) 
+		if (f->type != type && f->type != filt) 
 			continue;
 		if ((fres = func_cmp(sa, f, sqlfname, nrargs )) != NULL) {
 			MT_lock_unset(&funcs->ht_lock);
@@ -748,7 +774,7 @@ sql_find_func(sql_allocator *sa, sql_schema *s, const char *sqlfname, int nrargs
 				for (; he; he = he->chain) {
 					sql_func *f = he->value;
 
-					if (f->type != type)
+					if (f->type != type && f->type != filt) 
 						continue;
 					if ((fres = func_cmp(sa, f, sqlfname, nrargs )) != NULL) {
 						MT_lock_unset(&s->funcs.set->ht_lock);
@@ -769,7 +795,7 @@ sql_find_func(sql_allocator *sa, sql_schema *s, const char *sqlfname, int nrargs
 				for (; n; n = n->next) {
 					sql_func *f = n->data;
 
-					if (f->type != type)
+					if (f->type != type && f->type != filt) 
 						continue;
 					if ((fres = func_cmp(sa, f, sqlfname, nrargs )) != NULL)
 						return fres;
@@ -800,7 +826,7 @@ sql_bind_member(sql_allocator *sa, sql_schema *s, const char *sqlfname, sql_subt
 	for (; n; n = n->next) {
 		sql_func *f = n->data;
 
-		if (!f->res)
+		if (!f->res && !IS_FILT(f))
 			continue;
 		if (strcmp(f->base.name, sqlfname) == 0) {
 			if (list_length(f->ops) == nrargs && is_subtypeof(tp, &((sql_arg *) f->ops->h->data)->type)) 
@@ -823,7 +849,7 @@ sql_bind_member(sql_allocator *sa, sql_schema *s, const char *sqlfname, sql_subt
 		for (; n; n = n->next) {
 			sql_func *f = n->data;
 
-			if (!f->res)
+			if (!f->res && !IS_FILT(f))
 				continue;
 			if (strcmp(f->base.name, sqlfname) == 0) {
 				if (list_length(f->ops) == nrargs && is_subtypeof(tp, &((sql_arg *) f->ops->h->data)->type)) 
@@ -870,12 +896,13 @@ sql_subfunc *
 sql_bind_func_(sql_allocator *sa, sql_schema *s, const char *sqlfname, list *ops, int type)
 {
 	node *n = funcs->h;
+	int filt = (type == F_FUNC)?F_FILT:type;
 
 	(void)s;
 	for (; n; n = n->next) {
 		sql_func *f = n->data;
 
-		if (f->type != type) 
+		if (f->type != type && f->type != filt) 
 			continue;
 		if (strcmp(f->base.name, sqlfname) == 0) {
 			if (list_cmp(f->ops, ops, (fcmp) &arg_subtype_cmp) == 0) 
@@ -888,7 +915,7 @@ sql_bind_func_(sql_allocator *sa, sql_schema *s, const char *sqlfname, list *ops
 		if (s->funcs.set) for (n=s->funcs.set->h; n; n = n->next) {
 			sql_func *f = n->data;
 
-			if (f->type != type) 
+			if (f->type != type && f->type != filt) 
 				continue;
 			if (strcmp(f->base.name, sqlfname) == 0) {
 				if (list_cmp(f->ops, ops, (fcmp) &arg_subtype_cmp) == 0) 
@@ -941,9 +968,9 @@ sql_bind_func_result_(sql_allocator *sa, sql_schema *s, const char *sqlfname, li
 		sql_func *f = n->data;
 		sql_arg *firstres = NULL;
 
-		if (!f->res)
+		if (!f->res && !IS_FILT(f))
 			continue;
-		firstres = f->res->h->data;
+		firstres = IS_FILT(f)?BIT:f->res->h->data;
 		if (strcmp(f->base.name, sqlfname) == 0 && (is_subtype(&firstres->type, res) || firstres->type.type->eclass == EC_ANY) && list_cmp(f->ops, ops, (fcmp) &arg_subtype_cmp) == 0) 
 			return sql_dup_subfunc(sa, f, ops, NULL);
 	}
@@ -953,9 +980,9 @@ sql_bind_func_result_(sql_allocator *sa, sql_schema *s, const char *sqlfname, li
 		sql_func *f = n->data;
 		sql_arg *firstres = NULL;
 
-		if (!f->res)
+		if (!f->res && !IS_FILT(f))
 			continue;
-		firstres = f->res->h->data;
+		firstres = IS_FILT(f)?BIT:f->res->h->data;
 		if (strcmp(f->base.name, sqlfname) == 0 && (is_subtype(&firstres->type, res) || firstres->type.type->eclass == EC_ANY) && list_cmp(f->ops, ops, (fcmp) &arg_subtype_cmp) == 0) 
 			return sql_dup_subfunc(sa, f, ops, NULL);
 	}
@@ -1229,7 +1256,8 @@ sqltypeinit( sql_allocator *sa)
 	sql_type *ts[100];
 	sql_type **strings, **numerical;
 	sql_type **decimals, **floats, **dates, **end, **t;
-	sql_type *STR, *BTE, *SHT, *INT, *LNG, *OID, *BIT, *DBL, *WRD, *DEC;
+	sql_type *STR, *BTE, *SHT, *INT, *LNG, *OID, *FLT, *DBL, *DEC;
+	sql_type *WRD;
 #ifdef HAVE_HGE
 	sql_type *HGE = NULL;
 #endif
@@ -1266,13 +1294,13 @@ sqltypeinit( sql_allocator *sa)
 	BTE = *t++ = sql_create_type(sa, "TINYINT",   8, SCALE_FIX, 2, EC_NUM, "bte");
 	SHT = *t++ = sql_create_type(sa, "SMALLINT", 16, SCALE_FIX, 2, EC_NUM, "sht");
 	INT = *t++ = sql_create_type(sa, "INT",      32, SCALE_FIX, 2, EC_NUM, "int");
-#if SIZEOF_WRD == SIZEOF_INT
-	WRD = *t++ = sql_create_type(sa, "WRD", 32, SCALE_FIX, 2, EC_NUM, "wrd");
+#if SIZEOF_SIZE_T == SIZEOF_INT
+	WRD = *t++ = sql_create_type(sa, "WRD", 32, SCALE_FIX, 2, EC_NUM, "int");
 #endif
 	LargestINT =
 	LNG = *t++ = sql_create_type(sa, "BIGINT",   64, SCALE_FIX, 2, EC_NUM, "lng");
-#if SIZEOF_WRD == SIZEOF_LNG
-	WRD = *t++ = sql_create_type(sa, "WRD", 64, SCALE_FIX, 2, EC_NUM, "wrd");
+#if SIZEOF_SIZE_T == SIZEOF_LNG
+	WRD = *t++ = sql_create_type(sa, "WRD", 64, SCALE_FIX, 2, EC_NUM, "lng");
 #endif
 #ifdef HAVE_HGE
 	if (have_hge) {
@@ -1288,17 +1316,14 @@ sqltypeinit( sql_allocator *sa)
 	*t++ = sql_create_type(sa, "DECIMAL",  4, SCALE_FIX, 10, EC_DEC, "sht");
 	DEC =
 	*t++ = sql_create_type(sa, "DECIMAL",  9, SCALE_FIX, 10, EC_DEC, "int");
+	LargestDEC =
+	*t++ = sql_create_type(sa, "DECIMAL", 18, SCALE_FIX, 10, EC_DEC, "lng");
 #ifdef HAVE_HGE
 	if (have_hge) {
-		*t++ = sql_create_type(sa, "DECIMAL", 18, SCALE_FIX, 10, EC_DEC, "lng");
 		LargestDEC =
-		*t++ = sql_create_type(sa, "DECIMAL", 39, SCALE_FIX, 10, EC_DEC, "hge");
-	} else
-#endif
-	{
-		LargestDEC =
-		*t++ = sql_create_type(sa, "DECIMAL", 19, SCALE_FIX, 10, EC_DEC, "lng");
+		*t++ = sql_create_type(sa, "DECIMAL", 38, SCALE_FIX, 10, EC_DEC, "hge");
 	}
+#endif
 
 	/* float(n) (n indicates precision of atleast n digits) */
 	/* ie n <= 23 -> flt */
@@ -1307,7 +1332,7 @@ sqltypeinit( sql_allocator *sa)
 	/* this requires a type definition */
 
 	floats = t;
-	*t++ = sql_create_type(sa, "REAL", 24, SCALE_NOFIX, 2, EC_FLT, "flt");
+	FLT = *t++ = sql_create_type(sa, "REAL", 24, SCALE_NOFIX, 2, EC_FLT, "flt");
 	DBL = *t++ = sql_create_type(sa, "DOUBLE", 53, SCALE_NOFIX, 2, EC_FLT, "dbl");
 
 	dates = t;
@@ -1369,12 +1394,12 @@ sqltypeinit( sql_allocator *sa)
 
 	sql_create_aggr(sa, "not_unique", "sql", "not_unique", OID, BIT);
 	/* well to be precise it does reduce and map */
-	sql_create_func(sa, "not_uniques", "sql", "not_uniques", WRD, NULL, OID, SCALE_NONE);
+	sql_create_func(sa, "not_uniques", "sql", "not_uniques", LNG, NULL, OID, SCALE_NONE);
 	sql_create_func(sa, "not_uniques", "sql", "not_uniques", OID, NULL, OID, SCALE_NONE);
 
 	/* functions needed for all types */
-	sql_create_func(sa, "hash", "mkey", "hash", ANY, NULL, WRD, SCALE_FIX);
-	sql_create_func3(sa, "rotate_xor_hash", "calc", "rotate_xor_hash", WRD, INT, ANY, WRD, SCALE_NONE);
+	sql_create_func(sa, "hash", "mkey", "hash", ANY, NULL, LNG, SCALE_FIX);
+	sql_create_func3(sa, "rotate_xor_hash", "calc", "rotate_xor_hash", LNG, INT, ANY, LNG, SCALE_NONE);
 	sql_create_func(sa, "=", "calc", "=", ANY, ANY, BIT, SCALE_FIX);
 	sql_create_func(sa, "<>", "calc", "!=", ANY, ANY, BIT, SCALE_FIX);
 	sql_create_func(sa, "isnull", "calc", "isnil", ANY, NULL, BIT, SCALE_FIX);
@@ -1383,8 +1408,11 @@ sqltypeinit( sql_allocator *sa)
 	sql_create_func(sa, "<", "calc", "<", ANY, ANY, BIT, SCALE_FIX);
 	sql_create_func(sa, "<=", "calc", "<=", ANY, ANY, BIT, SCALE_FIX);
 	sql_create_aggr(sa, "zero_or_one", "sql", "zero_or_one", ANY, ANY);
+	sql_create_aggr(sa, "all", "sql", "all", ANY, ANY);
 	sql_create_aggr(sa, "exist", "aggr", "exist", ANY, BIT);
 	sql_create_aggr(sa, "not_exist", "aggr", "not_exist", ANY, BIT);
+	sql_create_func(sa, "sql_exists", "aggr", "exist", ANY, NULL, BIT, SCALE_NONE);
+	sql_create_func(sa, "sql_not_exists", "aggr", "not_exist", ANY, NULL, BIT, SCALE_NONE);
 	/* needed for relational version */
 	sql_create_func(sa, "in", "calc", "in", ANY, ANY, BIT, SCALE_NONE);
 	sql_create_func(sa, "identity", "calc", "identity", ANY, NULL, OID, SCALE_NONE);
@@ -1401,12 +1429,12 @@ sqltypeinit( sql_allocator *sa)
 	sql_create_aggr(sa, "sum", "aggr", "sum", BTE, LargestINT);
 	sql_create_aggr(sa, "sum", "aggr", "sum", SHT, LargestINT);
 	sql_create_aggr(sa, "sum", "aggr", "sum", INT, LargestINT);
-	sql_create_aggr(sa, "sum", "aggr", "sum", LNG, LargestINT);
+	//sql_create_aggr(sa, "sum", "aggr", "sum", LNG, LargestINT);
 #ifdef HAVE_HGE
 	if (have_hge)
 		sql_create_aggr(sa, "sum", "aggr", "sum", HGE, LargestINT);
 #endif
-	sql_create_aggr(sa, "sum", "aggr", "sum", WRD, WRD);
+	sql_create_aggr(sa, "sum", "aggr", "sum", LNG, LNG);
 
 	t = decimals; /* BTE */
 	sql_create_aggr(sa, "sum", "aggr", "sum", *(t), LargestDEC);
@@ -1432,7 +1460,7 @@ sqltypeinit( sql_allocator *sa)
 	if (HAVE_HGE)
 		sql_create_aggr(sa, "prod", "aggr", "prod", HGE, LargestINT);
 #endif
-	/*sql_create_aggr(sa, "prod", "aggr", "prod", WRD, WRD);*/
+	/*sql_create_aggr(sa, "prod", "aggr", "prod", LNG, LNG);*/
 
 	t = decimals; /* BTE */
 	sql_create_aggr(sa, "prod", "aggr", "prod", *(t), LargestDEC);
@@ -1449,8 +1477,11 @@ sqltypeinit( sql_allocator *sa)
 	}
 #endif
 
-	for (t = numerical; t < dates; t++) 
+	for (t = numerical; t < dates; t++) {
+		if (*t == WRD)
+			continue;
 		sql_create_func(sa, "mod", "calc", "%", *t, *t, *t, SCALE_FIX);
+	}
 
 	for (t = floats; t < dates; t++) {
 		sql_create_aggr(sa, "sum", "aggr", "sum", *t, *t);
@@ -1458,7 +1489,9 @@ sqltypeinit( sql_allocator *sa)
 	}
 	sql_create_aggr(sa, "sum", "aggr", "sum", MONINT, MONINT);
 	sql_create_aggr(sa, "sum", "aggr", "sum", SECINT, SECINT);
-	/*
+	/* do DBL first so that it is chosen as cast destination for
+	 * unknown types */
+	sql_create_aggr(sa, "avg", "aggr", "avg", DBL, DBL);
 	sql_create_aggr(sa, "avg", "aggr", "avg", BTE, DBL);
 	sql_create_aggr(sa, "avg", "aggr", "avg", SHT, DBL);
 	sql_create_aggr(sa, "avg", "aggr", "avg", INT, DBL);
@@ -1467,11 +1500,10 @@ sqltypeinit( sql_allocator *sa)
 	if (have_hge)
 		sql_create_aggr(sa, "avg", "aggr", "avg", HGE, DBL);
 #endif
-	*/
-	sql_create_aggr(sa, "avg", "aggr", "avg", DBL, DBL);
+	sql_create_aggr(sa, "avg", "aggr", "avg", FLT, DBL);
 
-	sql_create_aggr(sa, "count_no_nil", "aggr", "count_no_nil", NULL, WRD);
-	sql_create_aggr(sa, "count", "aggr", "count", NULL, WRD);
+	sql_create_aggr(sa, "count_no_nil", "aggr", "count_no_nil", NULL, LNG);
+	sql_create_aggr(sa, "count", "aggr", "count", NULL, LNG);
 
 	/* order based operators */
 	sql_create_analytic(sa, "diff", "sql", "diff", ANY, NULL, NULL, BIT, SCALE_NONE);
@@ -1501,16 +1533,35 @@ sqltypeinit( sql_allocator *sa)
 	/* allow smaller types for arguments of mul/div */
 	for (t = numerical, t++; t != decimals; t++) {
 		sql_type **u;
+		if (*t == WRD)
+			continue;
 		for (u = numerical, u++; u != decimals; u++) {
+			if (*u == WRD)
+				continue;
 			if (t != u && (*t)->localtype >  (*u)->localtype) {
 				sql_create_func(sa, "sql_mul", "calc", "*", *t, *u, *t, SCALE_MUL);
 				sql_create_func(sa, "sql_div", "calc", "/", *t, *u, *t, SCALE_DIV);
 			}
 		}
 	}
+	for (t = decimals, t++; t != floats; t++) {
+		sql_type **u;
+
+		for (u = decimals, u++; u != floats; u++) {
+			if (t != u && (*t)->localtype >  (*u)->localtype) {
+				sql_create_func(sa, "sql_mul", "calc", "*", *t, *u, *t, SCALE_MUL);
+				sql_create_func(sa, "sql_div", "calc", "/", *t, *u, *t, SCALE_DIV);
+			}
+		}
+	}
+
 	/* all numericals */
 	for (t = numerical; *t != TME; t++) {
-		sql_subtype *lt = sql_bind_localtype((*t)->base.name);
+		sql_subtype *lt;
+
+		if (*t == WRD)
+			continue;
+		lt = sql_bind_localtype((*t)->base.name);
 
 		sql_create_func(sa, "sql_sub", "calc", "-", *t, *t, *t, SCALE_FIX);
 		sql_create_func(sa, "sql_add", "calc", "+", *t, *t, *t, SCALE_FIX);
@@ -1545,6 +1596,8 @@ sqltypeinit( sql_allocator *sa)
 	for (t = decimals, t++; t != floats; t++) {
 		sql_type **u;
 		for (u = numerical; u != floats; u++) {
+			if (*u == WRD)
+				continue;
 			if (*u == OID)
 				continue;
 			if ((*t)->localtype >  (*u)->localtype) {
@@ -1560,8 +1613,13 @@ sqltypeinit( sql_allocator *sa)
 	for (t = numerical; t < end; t++) {
 		sql_type **u;
 
-		for (u = numerical; u < end; u++) 
+		if (*t == WRD)
+			continue;
+		for (u = numerical; u < end; u++) {
+			if (*u == WRD)
+				continue;
 			sql_create_func(sa, "scale_up", "calc", "*", *u, *t, *t, SCALE_NONE);
+		}
 	}
 
 	for (t = floats; t < dates; t++) {
@@ -1680,14 +1738,17 @@ sqltypeinit( sql_allocator *sa)
 		sql_create_func3(sa, "substring", "str", "substring", *t, INT, INT, *t, INOUT);
 		sql_create_func(sa, "substr", "str", "substring", *t, INT, *t, INOUT);
 		sql_create_func3(sa, "substr", "str", "substring", *t, INT, INT, *t, INOUT);
+		/*
 		sql_create_func(sa, "like", "algebra", "like", *t, *t, BIT, SCALE_NONE);
 		sql_create_func3(sa, "like", "algebra", "like", *t, *t, *t, BIT, SCALE_NONE);
 		sql_create_func(sa, "ilike", "algebra", "ilike", *t, *t, BIT, SCALE_NONE);
 		sql_create_func3(sa, "ilike", "algebra", "ilike", *t, *t, *t, BIT, SCALE_NONE);
+		*/
 		sql_create_func(sa, "not_like", "algebra", "not_like", *t, *t, BIT, SCALE_NONE);
 		sql_create_func3(sa, "not_like", "algebra", "not_like", *t, *t, *t, BIT, SCALE_NONE);
 		sql_create_func(sa, "not_ilike", "algebra", "not_ilike", *t, *t, BIT, SCALE_NONE);
 		sql_create_func3(sa, "not_ilike", "algebra", "not_ilike", *t, *t, *t, BIT, SCALE_NONE);
+
 		sql_create_func(sa, "patindex", "pcre", "patindex", *t, *t, INT, SCALE_NONE);
 		sql_create_func(sa, "truncate", "str", "stringleft", *t, INT, *t, SCALE_NONE);
 		sql_create_func(sa, "concat", "calc", "+", *t, *t, *t, DIGITS_ADD);
@@ -1740,9 +1801,9 @@ sqltypeinit( sql_allocator *sa)
 				sres, FALSE, F_FUNC, SCALE_FIX);
 	}
 	sres = create_arg(sa, NULL, sql_create_subtype(sa, TABLE, 0, 0), ARG_OUT); 
-	/* copyfrom fname (arg 10) */
+	/* copyfrom fname (arg 11) */
 	f=sql_create_func_(sa, "copyfrom", "sql", "copy_from",
-	 	list_append( list_append( list_append( list_append( list_append(list_append (list_append (list_append(list_append(list_append(sa_list(sa),
+	 	list_append( list_append( list_append( list_append( list_append( list_append(list_append (list_append (list_append(list_append(list_append(sa_list(sa),
 			create_arg(sa, NULL, sql_create_subtype(sa, STR, 0, 0), ARG_IN)), 
 			create_arg(sa, NULL, sql_create_subtype(sa, STR, 0, 0), ARG_IN)), 
 			create_arg(sa, NULL, sql_create_subtype(sa, STR, 0, 0), ARG_IN)), 
@@ -1752,9 +1813,14 @@ sqltypeinit( sql_allocator *sa)
 			create_arg(sa, NULL, sql_create_subtype(sa, LNG, 0, 0), ARG_IN)), 
 			create_arg(sa, NULL, sql_create_subtype(sa, LNG, 0, 0), ARG_IN)), 
 			create_arg(sa, NULL, sql_create_subtype(sa, INT, 0, 0), ARG_IN)),
+			create_arg(sa, NULL, sql_create_subtype(sa, INT, 0, 0), ARG_IN)),
 			create_arg(sa, NULL, sql_create_subtype(sa, STR, 0, 0), ARG_IN)), sres, FALSE, F_UNION, SCALE_FIX);
 	f->varres = 1;
 
+	f=sql_create_func_(sa, "append", "sql", "append", sa_list(sa), sres, FALSE, F_UNION, SCALE_FIX);
+	f->varres = 1;
+	f->vararg = 1;
+
 	/* bincopyfrom */
 	f = sql_create_func_(sa, "copyfrom", "sql", "importTable",
 	 	list_append(list_append(sa_list(sa), 
