@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #include "monetdb_config.h"
@@ -15,6 +15,7 @@
 #include "sql_privileges.h"
 #include "rel_optimizer.h"
 #include "rel_dump.h"
+#include "rel_psm.h"
 #include "sql_symbol.h"
 
 static sql_exp *
@@ -84,25 +85,25 @@ rel_insert_hash_idx(mvc *sql, sql_idx *i, sql_rel *inserts)
 {
 	char *iname = sa_strconcat( sql->sa, "%", i->base.name);
 	node *m;
-	sql_subtype *it, *wrd;
-	int bits = 1 + ((sizeof(wrd)*8)-1)/(list_length(i->columns)+1);
+	sql_subtype *it, *lng;
+	int bits = 1 + ((sizeof(lng)*8)-1)/(list_length(i->columns)+1);
 	sql_exp *h = NULL;
 
 	if (list_length(i->columns) <= 1 || i->type == no_idx) {
 		/* dummy append */
-		append(get_inserts(inserts), exp_label(sql->sa, exp_atom_wrd(sql->sa, 0), ++sql->label));
+		append(get_inserts(inserts), exp_label(sql->sa, exp_atom_lng(sql->sa, 0), ++sql->label));
 		return inserts;
 	}
 
 	it = sql_bind_localtype("int");
-	wrd = sql_bind_localtype("wrd");
+	lng = sql_bind_localtype("lng");
 	for (m = i->columns->h; m; m = m->next) {
 		sql_kc *c = m->data;
 		sql_exp *e = list_fetch(get_inserts(inserts), c->c->colnr);
 
 		if (h && i->type == hash_idx)  { 
 			list *exps = new_exp_list(sql->sa);
-			sql_subfunc *xor = sql_bind_func_result3(sql->sa, sql->session->schema, "rotate_xor_hash", wrd, it, &c->c->type, wrd);
+			sql_subfunc *xor = sql_bind_func_result3(sql->sa, sql->session->schema, "rotate_xor_hash", lng, it, &c->c->type, lng);
 
 			append(exps, h);
 			append(exps, exp_atom_int(sql->sa, bits));
@@ -110,15 +111,15 @@ rel_insert_hash_idx(mvc *sql, sql_idx *i, sql_rel *inserts)
 			h = exp_op(sql->sa, exps, xor);
 		} else if (h)  { /* order preserving hash */
 			sql_exp *h2;
-			sql_subfunc *lsh = sql_bind_func_result(sql->sa, sql->session->schema, "left_shift", wrd, it, wrd);
-			sql_subfunc *lor = sql_bind_func_result(sql->sa, sql->session->schema, "bit_or", wrd, wrd, wrd);
-			sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, wrd);
+			sql_subfunc *lsh = sql_bind_func_result(sql->sa, sql->session->schema, "left_shift", lng, it, lng);
+			sql_subfunc *lor = sql_bind_func_result(sql->sa, sql->session->schema, "bit_or", lng, lng, lng);
+			sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, lng);
 
 			h = exp_binop(sql->sa, h, exp_atom_int(sql->sa, bits), lsh); 
 			h2 = exp_unop(sql->sa, e, hf);
 			h = exp_binop(sql->sa, h, h2, lor);
 		} else {
-			sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, wrd);
+			sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, lng);
 			h = exp_unop(sql->sa, e, hf);
 			if (i->type == oph_idx) 
 				break;
@@ -210,6 +211,7 @@ rel_insert_join_idx(mvc *sql, sql_idx *i, sql_rel *inserts)
 		rel_destroy(ins);
 		rt = inserts->r = rel_setop(sql->sa, _nlls, nnlls, op_union );
 		rt->exps = rel_projections(sql, nnlls, NULL, 1, 1);
+		set_processed(rt);
 	} else {
 		inserts->r = nnlls;
 	}
@@ -490,8 +492,16 @@ insert_into(mvc *sql, dlist *qname, dlist *columns, symbol *val_or_q)
 						sql_column *c = m->data;
 						sql_rel *r = NULL;
 						sql_exp *ins = insert_value(sql, c, &r, n->data.sym);
-						if (!ins || r)
+						if (!ins) 
 							return NULL;
+						if (r && inner)
+							inner = rel_crossproduct(sql->sa, inner, r, op_join);
+						else if (r) 
+							inner = r;
+						if (inner && !ins->name) {
+							exp_label(sql->sa, ins, ++sql->label);
+							ins = exp_column(sql->sa, exp_relname(ins), exp_name(ins), exp_subtype(ins), ins->card, has_nil(ins), is_intern(ins));
+						}
 						list_append(vals_list, ins);
 					}
 				} else {
@@ -503,7 +513,7 @@ insert_into(mvc *sql, dlist *qname, dlist *columns, symbol *val_or_q)
 						if (!ins)
 							return NULL;
 						if (r && inner)
-							inner = rel_crossproduct(sql->sa, inner,r, op_join);
+							inner = rel_crossproduct(sql->sa, inner, r, op_join);
 						else if (r) 
 							inner = r;
 						if (!ins->name)
@@ -562,15 +572,15 @@ rel_update_hash_idx(mvc *sql, sql_idx *i, sql_rel *updates)
 {
 	char *iname = sa_strconcat( sql->sa, "%", i->base.name);
 	node *m;
-	sql_subtype *it, *wrd = 0; /* is not set in first if below */
-	int bits = 1 + ((sizeof(wrd)*8)-1)/(list_length(i->columns)+1);
+	sql_subtype *it, *lng = 0; /* is not set in first if below */
+	int bits = 1 + ((sizeof(lng)*8)-1)/(list_length(i->columns)+1);
 	sql_exp *h = NULL;
 
 	if (list_length(i->columns) <= 1 || i->type == no_idx) {
-		h = exp_label(sql->sa, exp_atom_wrd(sql->sa, 0), ++sql->label);
+		h = exp_label(sql->sa, exp_atom_lng(sql->sa, 0), ++sql->label);
 	} else {
 		it = sql_bind_localtype("int");
-		wrd = sql_bind_localtype("wrd");
+		lng = sql_bind_localtype("lng");
 		for (m = i->columns->h; m; m = m->next) {
 			sql_kc *c = m->data;
 			sql_exp *e;
@@ -579,7 +589,7 @@ rel_update_hash_idx(mvc *sql, sql_idx *i, sql_rel *updates)
 			
 			if (h && i->type == hash_idx)  { 
 				list *exps = new_exp_list(sql->sa);
-				sql_subfunc *xor = sql_bind_func_result3(sql->sa, sql->session->schema, "rotate_xor_hash", wrd, it, &c->c->type, wrd);
+				sql_subfunc *xor = sql_bind_func_result3(sql->sa, sql->session->schema, "rotate_xor_hash", lng, it, &c->c->type, lng);
 	
 				append(exps, h);
 				append(exps, exp_atom_int(sql->sa, bits));
@@ -587,15 +597,15 @@ rel_update_hash_idx(mvc *sql, sql_idx *i, sql_rel *updates)
 				h = exp_op(sql->sa, exps, xor);
 			} else if (h)  { /* order preserving hash */
 				sql_exp *h2;
-				sql_subfunc *lsh = sql_bind_func_result(sql->sa, sql->session->schema, "left_shift", wrd, it, wrd);
-				sql_subfunc *lor = sql_bind_func_result(sql->sa, sql->session->schema, "bit_or", wrd, wrd, wrd);
-				sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, wrd);
+				sql_subfunc *lsh = sql_bind_func_result(sql->sa, sql->session->schema, "left_shift", lng, it, lng);
+				sql_subfunc *lor = sql_bind_func_result(sql->sa, sql->session->schema, "bit_or", lng, lng, lng);
+				sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, lng);
 	
 				h = exp_binop(sql->sa, h, exp_atom_int(sql->sa, bits), lsh); 
 				h2 = exp_unop(sql->sa, e, hf);
 				h = exp_binop(sql->sa, h, h2, lor);
 			} else {
-				sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, wrd);
+				sql_subfunc *hf = sql_bind_func_result(sql->sa, sql->session->schema, "hash", &c->c->type, NULL, lng);
 				h = exp_unop(sql->sa, e, hf);
 				if (i->type == oph_idx) 
 					break;
@@ -608,7 +618,7 @@ rel_update_hash_idx(mvc *sql, sql_idx *i, sql_rel *updates)
 
 	if (!updates->exps)
 		updates->exps = new_exp_list(sql->sa);
-	append(updates->exps, exp_column(sql->sa, i->t->base.name, iname, wrd, CARD_MULTI, 0, 0));
+	append(updates->exps, exp_column(sql->sa, i->t->base.name, iname, lng, CARD_MULTI, 0, 0));
 	return updates;
 }
 
@@ -720,6 +730,7 @@ rel_update_join_idx(mvc *sql, sql_idx *i, sql_rel *updates)
 		rel_destroy(ups);
 		rt = updates->r = rel_setop(sql->sa, _nlls, nnlls, op_union );
 		rt->exps = rel_projections(sql, nnlls, NULL, 1, 1);
+		set_processed(rt);
 	} else {
 		updates->r = nnlls;
 	}
@@ -837,13 +848,12 @@ update_check_column(mvc *sql, sql_table *t, sql_column *c, sql_exp *v, sql_rel *
 }
 
 static sql_rel *
-update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_where)
+update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_from, symbol *opt_where)
 {
 	char *sname = qname_schema(qname);
 	char *tname = qname_table(qname);
 	sql_schema *s = NULL;
 	sql_table *t = NULL;
-	sql_rel *bt = NULL;
 
 	if (sname && !(s=mvc_bind_schema(sql,sname))) {
 		(void) sql_error(sql, 02, "3F000!UPDATE: no such schema '%s'", sname);
@@ -866,7 +876,9 @@ update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_where)
 		list *exps;
 		dnode *n;
 		const char *rname = NULL;
+		sql_rel *res = NULL, *bt = rel_basetable(sql, t, t->base.name);
 
+		res = bt;
 #if 0
 			dlist *selection = dlist_create(sql->sa);
 			dlist *from_list = dlist_create(sql->sa);
@@ -912,6 +924,21 @@ update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_where)
 		}
 #endif
 
+		if (opt_from) {
+			dlist *fl = opt_from->data.lval;
+			dnode *n = NULL;
+			sql_rel *fnd = NULL;
+
+			for (n = fl->h; n && res; n = n->next) {
+				fnd = table_ref(sql, NULL, n->data.sym, 0);
+				if (fnd)
+					res = rel_crossproduct(sql->sa, res, fnd, op_join);
+				else
+					res = fnd;
+			}
+			if (!res) 
+				return NULL;
+		}
 		if (opt_where) {
 			int status = sql->session->status;
 	
@@ -920,21 +947,18 @@ update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_where)
 			r = rel_logical_exp(sql, NULL, opt_where, sql_where);
 			if (r) { /* simple predicate which is not using the to 
 				    be updated table. We add a select all */
-
-				sql_rel *l = bt = rel_basetable(sql, t, t->base.name );
-				r = rel_crossproduct(sql->sa, l, r, op_semi);
+				r = rel_crossproduct(sql->sa, res, r, op_semi);
 			} else {
 				sql->errstr[0] = 0;
 				sql->session->status = status;
-				bt = r = rel_basetable(sql, t, t->base.name );
-				r = rel_logical_exp(sql, r, opt_where, sql_where);
-				if (r && is_join(r->op))
+				r = rel_logical_exp(sql, res, opt_where, sql_where);
+				if (!opt_from && r && is_join(r->op))
 					r->op = op_semi;
 			}
 			if (!r) 
 				return NULL;
 		} else {	/* update all */
-			bt = r = rel_basetable(sql, t, t->base.name );
+			r = res;
 		}
 	
 		/* first create the project */
@@ -1037,7 +1061,6 @@ update_table(mvc *sql, dlist *qname, dlist *assignmentlist, symbol *opt_where)
 					return NULL;
 				}
 				list_append(exps, exp_column(sql->sa, t->base.name, cname, &c->type, CARD_MULTI, 0, 0));
-				assert(!updates[c->colnr]);
 				exp_setname(sql->sa, v, c->t->base.name, c->base.name);
 				updates[c->colnr] = v;
 			}
@@ -1135,6 +1158,19 @@ table_column_types(sql_allocator *sa, sql_table *t)
 	return types;
 }
 
+static list *
+table_column_names(sql_allocator *sa, sql_table *t)
+{
+	node *n;
+	list *types = sa_list(sa);
+
+	if (t->columns.set) for (n = t->columns.set->h; n; n = n->next) {
+		sql_column *c = n->data;
+		append(types, &c->base.name);
+	}
+	return types;
+}
+
 static sql_rel *
 rel_import(mvc *sql, sql_table *t, char *tsep, char *rsep, char *ssep, char *ns, char *filename, lng nr, lng offset, int locked, int best_effort, dlist *fwf_widths)
 {
@@ -1144,9 +1180,8 @@ rel_import(mvc *sql, sql_table *t, char *tsep, char *rsep, char *ssep, char *ns,
 	sql_subtype tpe;
 	sql_exp *import;
 	sql_schema *sys = mvc_bind_schema(sql, "sys");
-	sql_subfunc *f = sql_find_func(sql->sa, sys, "copyfrom", 10, F_UNION, NULL);
+	sql_subfunc *f = sql_find_func(sql->sa, sys, "copyfrom", 11, F_UNION, NULL);
 	char* fwf_string = NULL;
-
 	if (!f) /* we do expect copyfrom to be there */
 		return NULL;
 	f->res = table_column_types(sql->sa, t);
@@ -1161,7 +1196,7 @@ rel_import(mvc *sql, sql_table *t, char *tsep, char *rsep, char *ssep, char *ns,
 	if (fwf_widths && dlist_length(fwf_widths) > 0) {
 		dnode *dn;
 		int ncol = 0;
-		char* fwf_string_cur = fwf_string = GDKmalloc(20 * dlist_length(fwf_widths) + 1); // a 64 bit int needs 19 characters in decimal representation plus the separator
+		char *fwf_string_cur = fwf_string = sa_alloc(sql->sa, 20 * dlist_length(fwf_widths) + 1); // a 64 bit int needs 19 characters in decimal representation plus the separator
 
 		if (!fwf_string)
 			return NULL;
@@ -1179,15 +1214,15 @@ rel_import(mvc *sql, sql_table *t, char *tsep, char *rsep, char *ssep, char *ns,
 	append( args, exp_atom_str(sql->sa, filename, &tpe)); 
 	import = exp_op(sql->sa,  
 		append(
+			append(
 				append(
 					append(
-						append(
-							append( args,
-								exp_atom_lng(sql->sa, nr)),
-								exp_atom_lng(sql->sa, offset)),
-								exp_atom_int(sql->sa, locked)),
-								exp_atom_int(sql->sa, best_effort)),
-								exp_atom_str(sql->sa, fwf_string, &tpe)), f);
+						append( args,
+							exp_atom_lng(sql->sa, nr)),
+							exp_atom_lng(sql->sa, offset)),
+							exp_atom_int(sql->sa, locked)),
+							exp_atom_int(sql->sa, best_effort)),
+							exp_atom_str(sql->sa, fwf_string, &tpe)), f);
 
 	exps = new_exp_list(sql->sa);
 	for (n = t->columns.set->h; n; n = n->next) {
@@ -1215,7 +1250,6 @@ copyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, dlist *headers, d
 	lng offset = (nr_offset)?nr_offset->h->next->data.l_val:0;
 	list *collist;
 	int reorder = 0;
-
 	assert(!nr_offset || nr_offset->h->type == type_lng);
 	assert(!nr_offset || nr_offset->h->next->type == type_lng);
 	if (sname && !(s=mvc_bind_schema(sql, sname))) {
@@ -1251,6 +1285,8 @@ copyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, dlist *headers, d
 	}
 	/* lock the store, for single user/transaction */
 	if (locked) { 
+		if (headers)
+			return sql_error(sql, 02, "COPY INTO .. LOCKED: not allowed with column lists");
 		store_lock();
 		while (store_nr_active > 1) {
 			store_unlock();
@@ -1264,8 +1300,12 @@ copyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, dlist *headers, d
 	collist = check_table_columns(sql, t, columns, "COPY", tname);
 	if (!collist)
 		return NULL;
-	/* if collist has skip and different order (or format specification) use intermediate table */
+	/* If we have a header specification use intermediate table, for
+	 * column specification other then the default list we need to reorder
+	 */
 	nt = t;
+	if (headers || collist != t->columns.set) 
+		reorder = 1;
 	if (headers) {
 		int has_formats = 0;
 		dnode *n;
@@ -1321,8 +1361,10 @@ copyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, dlist *headers, d
 
 			if (!rel)
 				rel = nrel;
-			else
+			else {
 				rel = rel_setop(sql->sa, rel, nrel, op_union);
+				set_processed(rel);
+			}
 			if (!rel)
 				return rel;
 		}
@@ -1387,7 +1429,7 @@ copyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, dlist *headers, d
 }
 
 static sql_rel *
-bincopyfrom(mvc *sql, dlist *qname, dlist *files, int constraint)
+bincopyfrom(mvc *sql, dlist *qname, dlist *columns, dlist *files, int constraint)
 {
 	char *sname = qname_schema(qname);
 	char *tname = qname_table(qname);
@@ -1398,11 +1440,14 @@ bincopyfrom(mvc *sql, dlist *qname, dlist *files, int constraint)
 	node *n;
 	sql_rel *res;
 	list *exps, *args;
-	sql_subtype tpe;
+	sql_subtype strtpe;
 	sql_exp *import;
 	sql_schema *sys = mvc_bind_schema(sql, "sys");
 	sql_subfunc *f = sql_find_func(sql->sa, sys, "copyfrom", 2, F_UNION, NULL); 
+	list *collist;
+	int i;
 
+	assert(f);
 	if (!copy_allowed(sql, 1)) {
 		(void) sql_error(sql, 02, "COPY INTO: insufficient privileges: "
 				"binary COPY INTO requires database administrator rights");
@@ -1427,18 +1472,37 @@ bincopyfrom(mvc *sql, dlist *qname, dlist *files, int constraint)
 	if (files == NULL)
 		return sql_error(sql, 02, "COPY INTO: must specify files");
 
+	collist = check_table_columns(sql, t, columns, "COPY BINARY", tname);
+	if (!collist)
+		return NULL;
+
 	f->res = table_column_types(sql->sa, t);
- 	sql_find_subtype(&tpe, "varchar", 0, 0);
-	args = append( append( new_exp_list(sql->sa), 
-		exp_atom_str(sql->sa, t->s?t->s->base.name:NULL, &tpe)), 
-		exp_atom_str(sql->sa, t->base.name, &tpe));
+ 	sql_find_subtype(&strtpe, "varchar", 0, 0);
+	args = append( append( new_exp_list(sql->sa),
+		exp_atom_str(sql->sa, t->s?t->s->base.name:NULL, &strtpe)), 
+		exp_atom_str(sql->sa, t->base.name, &strtpe));
 
-	for (dn = files->h; dn; dn = dn->next) {
-		append(args, exp_atom_str(sql->sa, dn->data.sval, &tpe)); 
-
-		/* extend the bincopyfrom, with extra args and types */
+	// create the list of files that is passed to the function as parameter
+	for(i = 0; i < list_length(t->columns.set); i++) {
+		// we have one file per column, however, because we have column selection that file might be NULL
+		// first, check if this column number is present in the passed in the parameters
+		int found = 0;
+		dn = files->h;
+		for (n = collist->h; n && dn; n = n->next, dn = dn->next) {
+			sql_column *c = n->data;
+			if (i == c->colnr) {
+				// this column number was present in the input arguments; pass in the file name
+				append(args, exp_atom_str(sql->sa, dn->data.sval, &strtpe)); 
+				found = 1;
+				break;
+			}
+		}
+		if (!found) {
+			// this column was not present in the input arguments; pass in NULL
+			append(args, exp_atom_str(sql->sa, NULL, &strtpe)); 
+		}
 	}
-	
+
 	import = exp_op(sql->sa,  args, f); 
 
 	exps = new_exp_list(sql->sa);
@@ -1453,6 +1517,63 @@ bincopyfrom(mvc *sql, dlist *qname, dlist *files, int constraint)
 	return res;
 }
 
+
+static sql_rel *
+copyfromloader(mvc *sql, dlist *qname, symbol *fcall)
+{
+	char *sname = qname_schema(qname);
+	char *tname = qname_table(qname);
+
+	sql_schema *s = NULL;
+	sql_table *t = NULL;
+
+	node *n;
+	sql_rel res_obj ;
+	sql_rel *res = &res_obj;
+	list *exps = new_exp_list(sql->sa); //, *args = NULL;
+	sql_exp *import;
+	exp_kind ek = {type_value, card_loader, FALSE};
+
+	if (!copy_allowed(sql, 1)) {
+		(void) sql_error(sql, 02, "COPY INTO: insufficient privileges: "
+				"binary COPY INTO requires database administrator rights");
+		return NULL;
+	}
+
+	if (sname && !(s=mvc_bind_schema(sql, sname))) {
+		(void) sql_error(sql, 02, "3F000!COPY INTO: no such schema '%s'", sname);
+		return NULL;
+	}
+	if (!s)
+		s = cur_schema(sql);
+	t = mvc_bind_table(sql, s, tname);
+	if (!t && !sname) {
+		s = tmp_schema(sql);
+		t = mvc_bind_table(sql, s, tname);
+		if (!t)
+			t = stack_find_table(sql, tname);
+	}
+	if (insert_allowed(sql, t, tname, "COPY INTO", "copy into") == NULL) {
+		return NULL;
+	}
+
+	import = rel_value_exp(sql, &res, fcall, sql_sel, ek);
+	if (!import) {
+		return NULL;
+	}
+	((sql_subfunc*) import->f)->res = table_column_types(sql->sa, t);
+	((sql_subfunc*) import->f)->colnames = table_column_names(sql->sa, t);
+
+	for (n = t->columns.set->h; n; n = n->next) {
+		sql_column *c = n->data;
+		append(exps, exp_column(sql->sa, t->base.name, c->base.name, &c->type, CARD_MULTI, c->null, 0));
+	}
+
+	res = rel_table_func(sql->sa, NULL, import, exps, 1);
+	return  rel_insert_table(sql, t, t->base.name, res);
+}
+
+
 static sql_rel *
 rel_output(mvc *sql, sql_rel *l, sql_exp *sep, sql_exp *rsep, sql_exp *ssep, sql_exp *null_string, sql_exp *file) 
 {
@@ -1527,7 +1648,7 @@ rel_parse_val(mvc *m, char *query, char emode)
 
 	m->caching = 0;
 	m->emode = emode;
-
+	// FIXME unchecked_malloc GDKmalloc can return NULL
 	b = (buffer*)GDKmalloc(sizeof(buffer));
 	n = GDKmalloc(len + 1 + 1);
 	strncpy(n, query, len);
@@ -1611,7 +1732,15 @@ rel_updates(mvc *sql, symbol *s)
 	{
 		dlist *l = s->data.lval;
 
-		ret = bincopyfrom(sql, l->h->data.lval, l->h->next->data.lval, l->h->next->next->data.i_val);
+		ret = bincopyfrom(sql, l->h->data.lval, l->h->next->data.lval, l->h->next->next->data.lval, l->h->next->next->next->data.i_val);
+		sql->type = Q_UPDATE;
+	}
+		break;
+	case SQL_COPYLOADER:
+	{
+		dlist *l = s->data.lval;
+
+		ret = copyfromloader(sql, l->h->data.lval, l->h->next->data.sym);
 		sql->type = Q_UPDATE;
 	}
 		break;
@@ -1635,7 +1764,7 @@ rel_updates(mvc *sql, symbol *s)
 	{
 		dlist *l = s->data.lval;
 
-		ret = update_table(sql, l->h->data.lval, l->h->next->data.lval, l->h->next->next->data.sym);
+		ret = update_table(sql, l->h->data.lval, l->h->next->data.lval, l->h->next->next->data.sym, l->h->next->next->next->data.sym);
 		sql->type = Q_UPDATE;
 	}
 		break;
