@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 /*
@@ -45,6 +45,7 @@
 qc *
 qc_create(int clientid, int seqnr)
 {
+	// FIXME unchecked_malloc MNEW can return NULL
 	qc *r = MNEW(qc);
 	r->clientid = clientid;
 	r->id = seqnr;
@@ -236,9 +237,10 @@ qc_match(qc *cache, symbol *s, atom **params, int  plen, int key)
 }
 
 cq *
-qc_insert(qc *cache, sql_allocator *sa, sql_rel *r, symbol *s, atom **params, int paramlen, int key, int type, char *cmd)
+qc_insert(qc *cache, sql_allocator *sa, sql_rel *r, char *qname,  symbol *s, atom **params, int paramlen, int key, int type, char *cmd)
 {
 	int i, namelen;
+	// FIXME unchecked_malloc MNEW can return NULL
 	cq *n = MNEW(cq);
 
 	n->id = cache->id++;
@@ -267,11 +269,23 @@ qc_insert(qc *cache, sql_allocator *sa, sql_rel *r, symbol *s, atom **params, in
 	n->count = 1;
 	namelen = 5 + ((n->id+7)>>3) + ((cache->clientid+7)>>3);
 	n->name = sa_alloc(sa, namelen);
-	(void) snprintf(n->name, namelen, "s%d_%d", n->id, cache->clientid);
+	strcpy(n->name, qname);
 	cache->q = n;
 	return n;
 }
 
+int
+qc_isaquerytemplate(str name){
+	int i,j;
+	return sscanf(name, "s%d_%d", &i,&j) == 2;
+}
+
+int
+qc_isapreparedquerytemplate(str name){
+	int i,j;
+	return sscanf(name, "p%d_%d", &i,&j) == 2;
+}
+
 int
 qc_size(qc *cache)
 {
