@@ -3,7 +3,7 @@
  * License, v. 2.0.  If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  *
- * Copyright 1997 - July 2008 CWI, August 2008 - 2016 MonetDB B.V.
+ * Copyright 1997 - July 2008 CWI, August 2008 - 2017 MonetDB B.V.
  */
 
 #ifndef BAT_TABLE_H
@@ -14,6 +14,6 @@
 #include "bat_utils.h"
 
 /* initialize bat storage call back functions interface */
-extern int bat_table_init( table_functions *tf );
+extern void bat_table_init( table_functions *tf );
 
 #endif /*BAT_TABLE_H*/
